#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from taskflow.patterns import linear_flow
from taskflow.patterns import unordered_flow

from octavia.common import constants
from octavia.common import exceptions


from octavia.controller.worker.v1.flows import amphora_flows
from octavia.controller.worker.v1.flows import listener_flows
from octavia.controller.worker.v1.flows import member_flows
from octavia.controller.worker.v1.flows import pool_flows
from octavia.controller.worker.v1.tasks import compute_tasks
from octavia.controller.worker.v1.tasks import database_tasks
from octavia.controller.worker.v1.tasks import lifecycle_tasks
from octavia.controller.worker.v1.tasks import network_tasks
from octavia.db import api as db_apis
from octavia.db import repositories as repo

from a10_octavia.common import a10constants
from a10_octavia.controller.worker.flows import a10_l7policy_flows
from a10_octavia.controller.worker.flows import a10_listener_flows
from a10_octavia.controller.worker.flows import a10_pool_flows
from a10_octavia.controller.worker.flows import vthunder_flows
from a10_octavia.controller.worker.tasks import a10_compute_tasks
from a10_octavia.controller.worker.tasks import a10_database_tasks
from a10_octavia.controller.worker.tasks import a10_network_tasks
from a10_octavia.controller.worker.tasks import glm_tasks
from a10_octavia.controller.worker.tasks import nat_pool_tasks
from a10_octavia.controller.worker.tasks import virtual_server_tasks
from a10_octavia.controller.worker.tasks import vthunder_tasks
from a10_octavia.db import repositories as a10repo


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class LoadBalancerFlows(object):

    def __init__(self):
        self.amp_flows = amphora_flows.AmphoraFlows()
        self.listener_flows = listener_flows.ListenerFlows()
        self.pool_flows = pool_flows.PoolFlows()
        self.member_flows = member_flows.MemberFlows()
        self.vthunder_flows = vthunder_flows.VThunderFlows()
        self._listener_flows = a10_listener_flows.ListenerFlows()
        self._l7policy_flows = a10_l7policy_flows.L7PolicyFlows()
        self._pool_flows = a10_pool_flows.PoolFlows()
        self._lb_repo = repo.LoadBalancerRepository()
        self._vthunder_repo = a10repo.VThunderRepository()
        self.loadbalancer_repo = a10repo.LoadBalancerRepository()

    def get_create_load_balancer_flow(self, load_balancer_id, topology, project_id,
                                      listeners=None):
        """Flow to create a load balancer"""

        f_name = constants.CREATE_LOADBALANCER_FLOW
        lb_create_flow = linear_flow.Flow(f_name)
        lb_create_flow.add(lifecycle_tasks.LoadBalancerIDToErrorOnRevertTask(
            requires=constants.LOADBALANCER_ID))
        lb_create_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))

        lb_create_flow.add(database_tasks.ReloadLoadBalancer(
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER))

        lb_create_flow.add(a10_database_tasks.CheckExistingVthunderTopology(
            requires=constants.LOADBALANCER,
            inject={"topology": topology}))

        # Attaching vThunder to LB in database
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            lb_create_flow.add(*self._create_active_standby_topology())
            LOG.info("TOPOLOGY === " + str(topology))
        elif topology == constants.TOPOLOGY_SINGLE:
            lb_create_flow.add(*self._create_single_topology())
            LOG.info("TOPOLOGY === " + str(topology))
        else:
            LOG.error("Unknown topology: %s.  Unable to build load balancer.",
                      topology)
            raise exceptions.InvalidTopology(topology=topology)

        # IMP: Now creating vThunder config here
        post_amp_prefix = constants.POST_LB_AMP_ASSOCIATION_SUBFLOW
        vthunder = self._vthunder_repo.get_vthunder_by_project_id(db_apis.get_session(),
                                                                  project_id)
        lb_create_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.FLAVOR_DATA))
        lb_create_flow.add(
            self.get_post_lb_vthunder_association_flow(
                post_amp_prefix, load_balancer_id, topology, vthunder,
                mark_active=(not listeners)))
        lb_create_flow.add(a10_database_tasks.CountLoadbalancersWithFlavor(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER),
            provides=a10constants.LB_COUNT_FLAVOR))
        lb_create_flow.add(vthunder_tasks.AllowL2DSR(
            requires=(constants.SUBNET, constants.AMPHORA,
                      a10constants.LB_COUNT_FLAVOR, constants.FLAVOR_DATA)))
        lb_create_flow.add(nat_pool_tasks.NatPoolCreate(
            requires=(constants.LOADBALANCER,
                      a10constants.VTHUNDER, constants.FLAVOR_DATA)))
        lb_create_flow.add(virtual_server_tasks.CreateVirtualServerTask(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER,
                      constants.FLAVOR_DATA)))
        lb_create_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        lb_create_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))

        return lb_create_flow

    def _create_single_topology(self):
        return (self.vthunder_flows.get_vthunder_for_lb_subflow(
            prefix=constants.ROLE_STANDALONE,
            role=constants.ROLE_MASTER),)

    def _create_active_standby_topology(
            self, lf_name=constants.CREATE_LOADBALANCER_FLOW):
        anti_affinity = CONF.a10_nova.enable_anti_affinity
        flows = []
        if anti_affinity:
            flows.append(
                a10_compute_tasks.NovaServerGroupCreate(
                    name=lf_name + '-' +
                    constants.CREATE_SERVER_GROUP_FLOW,
                    requires=(constants.LOADBALANCER_ID),
                    provides=constants.SERVER_GROUP_ID))

            flows.append(
                database_tasks.UpdateLBServerGroupInDB(
                    name=lf_name + '-' +
                    constants.UPDATE_LB_SERVERGROUPID_FLOW,
                    requires=(constants.LOADBALANCER_ID,
                              constants.SERVER_GROUP_ID)))

        f_name = constants.CREATE_LOADBALANCER_FLOW
        amps_flow = unordered_flow.Flow(f_name)
        master_amp_sf = self.vthunder_flows.get_vthunder_for_lb_subflow(
            prefix=constants.ROLE_MASTER, role=constants.ROLE_MASTER)

        backup_amp_sf = self.vthunder_flows.get_vthunder_for_lb_subflow(
            prefix=constants.ROLE_BACKUP, role=constants.ROLE_BACKUP)
        amps_flow.add(master_amp_sf, backup_amp_sf)

        return flows + [amps_flow]

    def get_post_lb_vthunder_association_flow(self, prefix, load_balancer_id, topology,
                                              vthunder, mark_active=True):
        """Flow to manage networking after lb creation"""
        sf_name = prefix + '-' + constants.POST_LB_AMP_ASSOCIATION_SUBFLOW
        post_create_lb_flow = linear_flow.Flow(sf_name)
        post_create_lb_flow.add(
            database_tasks.ReloadLoadBalancer(
                name=sf_name + '-' + constants.RELOAD_LB_AFTER_AMP_ASSOC,
                requires=constants.LOADBALANCER_ID,
                provides=constants.LOADBALANCER))
        # IMP: here we will inject network flow
        new_LB_net_subflow = self.get_new_lb_networking_subflow(topology, vthunder)
        post_create_lb_flow.add(new_LB_net_subflow)

        post_create_lb_flow.add(
            vthunder_tasks.CreateHealthMonitorOnVThunder(
                name=a10constants.CREATE_HEALTH_MONITOR_ON_VTHUNDER,
                requires=a10constants.VTHUNDER))
        if not vthunder and topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            vrrp_subflow = self.vthunder_flows.get_vrrp_subflow(prefix)
            post_create_lb_flow.add(vrrp_subflow)

        post_create_lb_flow.add(database_tasks.UpdateLoadbalancerInDB(
            requires=[constants.LOADBALANCER, constants.UPDATE_DICT]))

        if CONF.a10_global.handle_vrid:
            post_create_lb_flow.add(self.handle_vrid_for_loadbalancer_subflow())

        if mark_active:
            post_create_lb_flow.add(database_tasks.MarkLBActiveInDB(
                name=sf_name + '-' + constants.MARK_LB_ACTIVE_INDB,
                requires=constants.LOADBALANCER))
        return post_create_lb_flow

    def get_delete_load_balancer_flow(self, lb, deleteCompute, cascade):
        """Flow to delete load balancer"""

        store = {}
        delete_LB_flow = linear_flow.Flow(constants.DELETE_LOADBALANCER_FLOW)
        delete_LB_flow.add(lifecycle_tasks.LoadBalancerToErrorOnRevertTask(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))

        delete_LB_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=(constants.LOADBALANCER, a10constants.MASTER_AMPHORA_STATUS),
            provides=a10constants.VTHUNDER))
        if lb.topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            delete_LB_flow.add(a10_compute_tasks.CheckAmphoraStatus(
                name="check-master-amphora-status",
                requires=a10constants.VTHUNDER,
                provides=a10constants.MASTER_AMPHORA_STATUS))
            delete_LB_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
                name="get-vthunder-on-the-basis-of-master-amphora-status",
                requires=(constants.LOADBALANCER, a10constants.MASTER_AMPHORA_STATUS),
                provides=a10constants.VTHUNDER))
            delete_LB_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        delete_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            requires=a10constants.VTHUNDER,
            inject={"status": constants.PENDING_DELETE}))
        delete_LB_flow.add(compute_tasks.NovaServerGroupDelete(
            requires=constants.SERVER_GROUP_ID))
        delete_LB_flow.add(database_tasks.MarkLBAmphoraeHealthBusy(
            requires=constants.LOADBALANCER))
        if cascade:
            (pools_listeners_delete, store) = self._get_cascade_delete_pools_listeners_flow(lb)
            delete_LB_flow.add(pools_listeners_delete)
        delete_LB_flow.add(database_tasks.GetAmphoraeFromLoadbalancer(
            requires=constants.LOADBALANCER_ID,
            provides=constants.AMPHORA))
        delete_LB_flow.add(a10_database_tasks.GetLoadBalancerListForDeletion(
            requires=(a10constants.VTHUNDER, constants.LOADBALANCER),
            provides=a10constants.LOADBALANCERS_LIST))
        if lb.topology == "ACTIVE_STANDBY":
            delete_LB_flow.add(a10_database_tasks.GetBackupVThunderByLoadBalancer(
                requires=(constants.LOADBALANCER, a10constants.VTHUNDER),
                provides=a10constants.BACKUP_VTHUNDER))
            delete_LB_flow.add(a10_compute_tasks.CheckAmphoraStatus(
                name="check-master-amphora-status-after_getting-backup",
                requires=a10constants.VTHUNDER,
                provides=a10constants.MASTER_AMPHORA_STATUS))
            delete_LB_flow.add(a10_compute_tasks.CheckAmphoraStatus(
                name="check-backup-amphora-status",
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
                provides=a10constants.BACKUP_AMPHORA_STATUS))

        if not deleteCompute:
            delete_LB_flow.add(a10_database_tasks.GetMemberListByProjectID(
                requires=a10constants.VTHUNDER,
                provides=a10constants.MEMBER_LIST))
            delete_LB_flow.add(a10_network_tasks.CalculateDelta(
                requires=(constants.LOADBALANCER, a10constants.LOADBALANCERS_LIST,
                          a10constants.MEMBER_LIST),
                provides=constants.DELTAS))
            delete_LB_flow.add(a10_network_tasks.HandleNetworkDeltas(
                requires=constants.DELTAS, provides=constants.ADDED_PORTS))
            if lb.topology == "ACTIVE_STANDBY":
                delete_LB_flow.add(vthunder_tasks.VCSSyncWait(
                    name="wait-vcs-ready-before-master-reload",
                    requires=(a10constants.VTHUNDER,
                              a10constants.MASTER_AMPHORA_STATUS,
                              a10constants.BACKUP_AMPHORA_STATUS)))
            delete_LB_flow.add(vthunder_tasks.AmphoraePostNetworkUnplug(
                name=a10constants.AMPHORA_POST_NETWORK_UNPLUG,
                requires=(constants.LOADBALANCER, constants.ADDED_PORTS, a10constants.VTHUNDER)))
            delete_LB_flow.add(
                vthunder_tasks.VThunderComputeConnectivityWait(
                    name=a10constants.VTHUNDER_CONNECTIVITY_WAIT,
                    requires=(a10constants.VTHUNDER, constants.AMPHORA,
                              a10constants.MASTER_AMPHORA_STATUS)))
            if lb.topology == "ACTIVE_STANDBY":
                delete_LB_flow.add(
                    vthunder_tasks.VThunderComputeConnectivityWait(
                        name=a10constants.BACKUP_CONNECTIVITY_WAIT,
                        rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER,
                                a10constants.MASTER_AMPHORA_STATUS:
                                a10constants.BACKUP_AMPHORA_STATUS},
                        requires=(constants.AMPHORA)))
                delete_LB_flow.add(vthunder_tasks.VCSSyncWait(
                    name="vip-unplug-wait-vcs-ready",
                    requires=(a10constants.VTHUNDER,
                              a10constants.MASTER_AMPHORA_STATUS,
                              a10constants.BACKUP_AMPHORA_STATUS)))
                delete_LB_flow.add(vthunder_tasks.GetMasterVThunder(
                    name=a10constants.GET_VTHUNDER_MASTER,
                    requires=a10constants.VTHUNDER,
                    provides=a10constants.VTHUNDER))
        delete_LB_flow.add(a10_network_tasks.GetLBResourceSubnet(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.SUBNET))
        delete_LB_flow.add(
            a10_database_tasks.GetChildProjectsOfParentPartition(
                requires=[a10constants.VTHUNDER],
                rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
                provides=a10constants.PARTITION_PROJECT_LIST
            ))
        delete_LB_flow.add(
            a10_database_tasks.CountLoadbalancersInProjectBySubnet(
                requires=[constants.SUBNET, a10constants.PARTITION_PROJECT_LIST],
                provides=a10constants.LB_COUNT_SUBNET))
        if CONF.a10_global.handle_vrid:
            delete_LB_flow.add(self.get_delete_lb_vrid_subflow(deleteCompute))
        delete_LB_flow.add(a10_network_tasks.DeallocateVIP(
            requires=[constants.LOADBALANCER, a10constants.LB_COUNT_SUBNET]))
        delete_LB_flow.add(virtual_server_tasks.DeleteVirtualServerTask(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER)))
        delete_LB_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.FLAVOR_DATA))
        delete_LB_flow.add(a10_database_tasks.CountLoadbalancersWithFlavor(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER),
            provides=a10constants.LB_COUNT_FLAVOR))
        delete_LB_flow.add(vthunder_tasks.DeleteL2DSR(
            requires=(constants.SUBNET, constants.AMPHORA,
                      a10constants.LB_COUNT_FLAVOR,
                      a10constants.LB_COUNT_SUBNET,
                      constants.FLAVOR_DATA)))
        delete_LB_flow.add(nat_pool_tasks.NatPoolDelete(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER,
                      a10constants.LB_COUNT_FLAVOR, constants.FLAVOR_DATA)))
        if deleteCompute:
            delete_LB_flow.add(glm_tasks.RevokeFlexpoolLicense(
                name=a10constants.REVOKE_ACTIVE_VTHUNDER_LICENSE,
                requires=a10constants.VTHUNDER))
            if lb.topology == "ACTIVE_STANDBY":
                delete_LB_flow.add(glm_tasks.RevokeFlexpoolLicense(
                    name=a10constants.REVOKE_BACKUP_VTHUNDER_LICENSE,
                    rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}))
            delete_LB_flow.add(compute_tasks.DeleteAmphoraeOnLoadBalancer(
                requires=constants.LOADBALANCER))
        delete_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            name=a10constants.MARK_VTHUNDER_MASTER_DELETED_IN_DB,
            requires=a10constants.VTHUNDER,
            inject={"status": constants.DELETED}))
        delete_LB_flow.add(database_tasks.MarkLBAmphoraeDeletedInDB(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.DisableLBAmphoraeHealthMonitoring(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.MarkLBDeletedInDB(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.DecrementLoadBalancerQuota(
            requires=constants.LOADBALANCER))
        if not deleteCompute:
            delete_LB_flow.add(vthunder_tasks.WriteMemory(
                requires=a10constants.VTHUNDER))
        delete_LB_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            name=a10constants.SET_THUNDER_UPDATE_AT,
            requires=a10constants.VTHUNDER))
        if lb.topology == "ACTIVE_STANDBY":
            delete_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
                name=a10constants.MARK_VTHUNDER_BACKUP_DELETED_IN_DB,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
                inject={"status": constants.DELETED}))
            delete_LB_flow.add(a10_database_tasks.SetThunderUpdatedAt(
                name=a10constants.SET_THUNDER_BACKUP_UPDATE_AT,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}))
        return (delete_LB_flow, store)

    def get_new_lb_networking_subflow(self, topology, vthunder):
        """Subflow to setup networking for amphora"""
        new_LB_net_subflow = linear_flow.Flow(constants.
                                              LOADBALANCER_NETWORKING_SUBFLOW)
        new_LB_net_subflow.add(database_tasks.GetAmphoraeFromLoadbalancer(
            requires=constants.LOADBALANCER_ID,
            provides=constants.AMPHORA))
        new_LB_net_subflow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            name=a10constants.GET_VTHUNDER_FROM_LB,
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        new_LB_net_subflow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            name=a10constants.GET_VTHUNDER_BY_LB,
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if vthunder:
            new_LB_net_subflow.add(a10_database_tasks.GetLoadBalancerListByProjectID(
                requires=a10constants.VTHUNDER,
                provides=a10constants.LOADBALANCERS_LIST))
            new_LB_net_subflow.add(a10_database_tasks.GetMemberListByProjectID(
                requires=a10constants.VTHUNDER,
                provides=a10constants.MEMBER_LIST))
            new_LB_net_subflow.add(a10_network_tasks.CalculateDelta(
                requires=(constants.LOADBALANCER, a10constants.LOADBALANCERS_LIST,
                          a10constants.MEMBER_LIST),
                provides=constants.DELTAS))
            new_LB_net_subflow.add(a10_network_tasks.HandleNetworkDeltas(
                requires=constants.DELTAS, provides=constants.ADDED_PORTS))
            if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
                new_LB_net_subflow.add(vthunder_tasks.VCSSyncWait(
                    name="wait-vcs-ready-before-master-reload",
                    requires=a10constants.VTHUNDER))
                new_LB_net_subflow.add(vthunder_tasks.GetMasterVThunder(
                    name=a10constants.GET_MASTER_VTHUNDER,
                    requires=a10constants.VTHUNDER,
                    provides=a10constants.VTHUNDER))
            # managing interface additions here
            new_LB_net_subflow.add(vthunder_tasks.AmphoraePostVIPPlug(
                name=a10constants.AMPHORAE_POST_VIP_PLUG,
                requires=(constants.LOADBALANCER, a10constants.VTHUNDER,
                          constants.ADDED_PORTS)))
            new_LB_net_subflow.add(
                vthunder_tasks.VThunderComputeConnectivityWait(
                    name=a10constants.VTHUNDER_CONNECTIVITY_WAIT,
                    requires=(a10constants.VTHUNDER, constants.AMPHORA)))
            if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
                new_LB_net_subflow.add(
                    a10_database_tasks.GetBackupVThunderByLoadBalancer(
                        name=a10constants.GET_BACKUP_VTHUNDER_BY_LB,
                        requires=constants.LOADBALANCER,
                        provides=a10constants.BACKUP_VTHUNDER))
                new_LB_net_subflow.add(
                    vthunder_tasks.VThunderComputeConnectivityWait(
                        name=a10constants.BACKUP_CONNECTIVITY_WAIT,
                        rebind={
                            a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
                        requires=constants.AMPHORA))
                new_LB_net_subflow.add(vthunder_tasks.VCSSyncWait(
                    name="wait-vcs-ready-after-reload",
                    requires=a10constants.VTHUNDER))
                new_LB_net_subflow.add(vthunder_tasks.GetMasterVThunder(
                    name=a10constants.GET_VTHUNDER_MASTER,
                    requires=a10constants.VTHUNDER,
                    provides=a10constants.VTHUNDER))
            new_LB_net_subflow.add(vthunder_tasks.EnableInterface(
                name=a10constants.ENABLE_VTHUNDER_INTERFACE,
                requires=(a10constants.VTHUNDER, constants.LOADBALANCER,
                          constants.ADDED_PORTS)))
        else:
            new_LB_net_subflow.add(vthunder_tasks.VThunderComputeConnectivityWait(
                name=a10constants.VTHUNDER_CONNECTIVITY_WAIT,
                requires=(a10constants.VTHUNDER, constants.AMPHORA)))
        new_LB_net_subflow.add(vthunder_tasks.EnableInterface(
            name=a10constants.ENABLE_MASTER_VTHUNDER_INTERFACE,
            inject={constants.ADDED_PORTS: {}},
            requires=(a10constants.VTHUNDER, constants.LOADBALANCER, constants.ADDED_PORTS)))
        new_LB_net_subflow.add(a10_database_tasks.MarkVThunderStatusInDB(
            name=a10constants.MARK_VTHUNDER_MASTER_ACTIVE_IN_DB,
            requires=a10constants.VTHUNDER,
            inject={a10constants.STATUS: constants.ACTIVE}))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            new_LB_net_subflow.add(
                a10_database_tasks.GetBackupVThunderByLoadBalancer(
                    name=a10constants.BACKUP_VTHUNDER,
                    requires=constants.LOADBALANCER,
                    provides=a10constants.BACKUP_VTHUNDER))
            new_LB_net_subflow.add(vthunder_tasks.EnableInterface(
                name=a10constants.ENABLE_BACKUP_VTHUNDER_INTERFACE,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
                inject={constants.ADDED_PORTS: {}},
                requires=(constants.LOADBALANCER, constants.ADDED_PORTS)))
            new_LB_net_subflow.add(a10_database_tasks.MarkVThunderStatusInDB(
                name=a10constants.MARK_VTHUNDER_BACKUP_ACTIVE_IN_DB,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
                inject={a10constants.STATUS: constants.ACTIVE}))
        new_LB_net_subflow.add(a10_network_tasks.PlugVIP(
            requires=constants.LOADBALANCER,
            provides=constants.AMPS_DATA))
        new_LB_net_subflow.add(a10_network_tasks.ApplyQos(
            requires=(constants.LOADBALANCER, constants.AMPS_DATA, constants.UPDATE_DICT)))
        new_LB_net_subflow.add(database_tasks.UpdateAmphoraeVIPData(
            requires=constants.AMPS_DATA))
        new_LB_net_subflow.add(database_tasks.ReloadLoadBalancer(
            name=constants.RELOAD_LB_AFTER_PLUG_VIP,
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER))
        return new_LB_net_subflow

    def get_update_load_balancer_flow(self, topology):
        """Flow to update load balancer."""

        update_LB_flow = linear_flow.Flow(constants.UPDATE_LOADBALANCER_FLOW)
        update_LB_flow.add(lifecycle_tasks.LoadBalancerToErrorOnRevertTask(
            requires=constants.LOADBALANCER))
        update_LB_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))
        update_LB_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            update_LB_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        update_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            requires=a10constants.VTHUNDER,
            inject={"status": constants.PENDING_UPDATE}))
        update_LB_flow.add(vthunder_tasks.SetupDeviceNetworkMap(
            requires=a10constants.VTHUNDER,
            provides=a10constants.VTHUNDER))
        update_LB_flow.add(network_tasks.ApplyQos(
            requires=(constants.LOADBALANCER, constants.UPDATE_DICT)))
        # update_LB_flow.add(amphora_driver_tasks.ListenersUpdate(
        #    requires=[constants.LOADBALANCER, constants.LISTENERS]))
        # post_create_lb_flow.add(handle_vrid_for_loadbalancer_subflow())
        if CONF.a10_global.handle_vrid:
            update_LB_flow.add(self.handle_vrid_for_loadbalancer_subflow())
        update_LB_flow.add(database_tasks.GetAmphoraeFromLoadbalancer(
            requires=constants.LOADBALANCER_ID,
            provides=constants.AMPHORA))
        update_LB_flow.add(a10_database_tasks.GetLoadbalancersInProjectBySubnet(
            requires=[constants.SUBNET, a10constants.PARTITION_PROJECT_LIST],
            provides=a10constants.LOADBALANCERS_LIST))
        update_LB_flow.add(a10_database_tasks.CheckForL2DSRFlavor(
            rebind={a10constants.LB_RESOURCE: a10constants.LOADBALANCERS_LIST},
            provides=a10constants.L2DSR_FLAVOR))
        update_LB_flow.add(a10_database_tasks.CountMembersInProjectBySubnet(
            requires=[constants.SUBNET, a10constants.PARTITION_PROJECT_LIST],
            provides=a10constants.MEMBER_COUNT))
        update_LB_flow.add(vthunder_tasks.UpdateL2DSR(
            requires=(constants.SUBNET, constants.AMPHORA,
                      a10constants.MEMBER_COUNT, a10constants.L2DSR_FLAVOR)))
        update_LB_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.FLAVOR_DATA))
        update_LB_flow.add(virtual_server_tasks.UpdateVirtualServerTask(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER,
                      constants.FLAVOR_DATA, constants.UPDATE_DICT)))
        update_LB_flow.add(database_tasks.UpdateLoadbalancerInDB(
            requires=[constants.LOADBALANCER, constants.UPDATE_DICT]))
        update_LB_flow.add(database_tasks.MarkLBActiveInDB(
            requires=constants.LOADBALANCER))
        update_LB_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        update_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            name="pending_update_to_active",
            requires=a10constants.VTHUNDER,
            inject={"status": constants.ACTIVE}))
        update_LB_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))
        return update_LB_flow

    def get_update_rack_load_balancer_flow(self, vthunder_conf, device_dict, topology):
        """Flow to update load balancer."""

        update_LB_flow = linear_flow.Flow(constants.UPDATE_LOADBALANCER_FLOW)
        update_LB_flow.add(lifecycle_tasks.LoadBalancerToErrorOnRevertTask(
            requires=constants.LOADBALANCER))

        # device-name flavor support
        update_LB_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.FLAVOR_DATA))
        update_LB_flow.add(vthunder_tasks.GetVthunderConfByFlavor(
            inject={a10constants.VTHUNDER_CONFIG: vthunder_conf,
                    a10constants.DEVICE_CONFIG_DICT: device_dict},
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER_CONFIG,
                      a10constants.DEVICE_CONFIG_DICT, constants.FLAVOR_DATA),
            provides=(a10constants.VTHUNDER_CONFIG, a10constants.USE_DEVICE_FLAVOR)))

        update_LB_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            update_LB_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        update_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            requires=a10constants.VTHUNDER,
            inject={"status": constants.PENDING_UPDATE}))
        update_LB_flow.add(vthunder_tasks.SetupDeviceNetworkMap(
            requires=a10constants.VTHUNDER,
            provides=a10constants.VTHUNDER))
        update_LB_flow.add(network_tasks.ApplyQos(
            requires=(constants.LOADBALANCER, constants.UPDATE_DICT)))
        if CONF.a10_global.handle_vrid:
            update_LB_flow.add(self.handle_vrid_for_loadbalancer_subflow())

        update_LB_flow.add(virtual_server_tasks.UpdateVirtualServerTask(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER,
                      constants.FLAVOR_DATA, constants.UPDATE_DICT)))
        update_LB_flow.add(database_tasks.UpdateLoadbalancerInDB(
            requires=[constants.LOADBALANCER, constants.UPDATE_DICT]))
        if CONF.a10_global.network_type == 'vlan':
            update_LB_flow.add(vthunder_tasks.TagInterfaceForLB(
                requires=[constants.LOADBALANCER,
                          a10constants.VTHUNDER]))
        update_LB_flow.add(database_tasks.MarkLBActiveInDB(
            requires=constants.LOADBALANCER))
        update_LB_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        update_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            name="pending_update_to_active",
            requires=a10constants.VTHUNDER,
            inject={"status": constants.ACTIVE}))
        update_LB_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))
        return update_LB_flow

    def get_create_rack_vthunder_load_balancer_flow(
            self, vthunder_conf, device_dict, topology, listeners=None):
        """Flow to create rack load balancer"""

        f_name = constants.CREATE_LOADBALANCER_FLOW
        lb_create_flow = linear_flow.Flow(f_name)

        lb_create_flow.add(lifecycle_tasks.LoadBalancerIDToErrorOnRevertTask(
            requires=constants.LOADBALANCER_ID))
        lb_create_flow.add(database_tasks.ReloadLoadBalancer(
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER))

        # device-name flavor support
        lb_create_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.FLAVOR_DATA))
        lb_create_flow.add(vthunder_tasks.GetVthunderConfByFlavor(
            inject={a10constants.VTHUNDER_CONFIG: vthunder_conf,
                    a10constants.DEVICE_CONFIG_DICT: device_dict},
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER_CONFIG,
                      a10constants.DEVICE_CONFIG_DICT, constants.FLAVOR_DATA),
            provides=(a10constants.VTHUNDER_CONFIG, a10constants.USE_DEVICE_FLAVOR)))
        lb_create_flow.add(vthunder_tasks.HandleACOSPartitionChange(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER_CONFIG),
            provides=a10constants.VTHUNDER_CONFIG))
        lb_create_flow.add(
            a10_database_tasks.CheckExistingThunderToProjectMappedEntries(
                requires=(
                    constants.LOADBALANCER,
                    a10constants.VTHUNDER_CONFIG,
                    a10constants.USE_DEVICE_FLAVOR)))
        lb_create_flow.add(
            self.vthunder_flows.get_rack_vthunder_for_lb_subflow(
                vthunder_conf=a10constants.VTHUNDER_CONFIG,
                prefix=constants.ROLE_STANDALONE,
                role=constants.ROLE_STANDALONE))
        post_amp_prefix = constants.POST_LB_AMP_ASSOCIATION_SUBFLOW
        lb_create_flow.add(
            self.get_post_lb_rack_vthunder_association_flow(
                post_amp_prefix, topology, mark_active=(not listeners)))
        lb_create_flow.add(nat_pool_tasks.NatPoolCreate(
            requires=(constants.LOADBALANCER,
                      a10constants.VTHUNDER, constants.FLAVOR_DATA)))
        lb_create_flow.add(virtual_server_tasks.CreateVirtualServerTask(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER,
                      constants.FLAVOR_DATA),
            provides=a10constants.STATUS))
        lb_create_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        lb_create_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))
        return lb_create_flow

    def get_delete_rack_vthunder_load_balancer_flow(self, lb, cascade, vthunder_conf, device_dict):
        """Flow to delete rack load balancer"""

        store = {}
        delete_LB_flow = linear_flow.Flow(constants.DELETE_LOADBALANCER_FLOW)
        delete_LB_flow.add(lifecycle_tasks.LoadBalancerToErrorOnRevertTask(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        delete_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            requires=a10constants.VTHUNDER,
            inject={"status": constants.PENDING_DELETE}))
        delete_LB_flow.add(vthunder_tasks.SetupDeviceNetworkMap(
            requires=a10constants.VTHUNDER,
            provides=a10constants.VTHUNDER))
        delete_LB_flow.add(compute_tasks.NovaServerGroupDelete(
            requires=constants.SERVER_GROUP_ID))
        delete_LB_flow.add(database_tasks.MarkLBAmphoraeHealthBusy(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.FLAVOR_DATA))
        delete_LB_flow.add(vthunder_tasks.GetVthunderConfByFlavor(
            inject={a10constants.VTHUNDER_CONFIG: vthunder_conf,
                    a10constants.DEVICE_CONFIG_DICT: device_dict},
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER_CONFIG,
                      a10constants.DEVICE_CONFIG_DICT, constants.FLAVOR_DATA),
            provides=(a10constants.VTHUNDER_CONFIG, a10constants.USE_DEVICE_FLAVOR)))
        if cascade:
            (pools_listeners_delete, store) = self._get_cascade_delete_pools_listeners_flow(lb)
            delete_LB_flow.add(pools_listeners_delete)
        delete_LB_flow.add(a10_network_tasks.GetLBResourceSubnet(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.SUBNET))
        delete_LB_flow.add(
            a10_database_tasks.GetChildProjectsOfParentPartition(
                requires=[a10constants.VTHUNDER],
                rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
                provides=a10constants.PARTITION_PROJECT_LIST
            ))
        delete_LB_flow.add(
            a10_database_tasks.CountLoadbalancersInProjectBySubnet(
                requires=[
                    constants.SUBNET,
                    a10constants.PARTITION_PROJECT_LIST,
                    a10constants.USE_DEVICE_FLAVOR],
                provides=a10constants.LB_COUNT_SUBNET))
        delete_LB_flow.add(
            a10_database_tasks.CountLoadbalancersOnThunderBySubnet(
                requires=[a10constants.VTHUNDER, constants.SUBNET, a10constants.USE_DEVICE_FLAVOR],
                provides=a10constants.LB_COUNT_THUNDER))
        if CONF.a10_global.handle_vrid:
            delete_LB_flow.add(self.get_delete_rack_lb_vrid_subflow())
        delete_LB_flow.add(a10_network_tasks.DeallocateVIP(
            requires=[constants.LOADBALANCER, a10constants.LB_COUNT_SUBNET]))
        delete_LB_flow.add(virtual_server_tasks.DeleteVirtualServerTask(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER)))
        delete_LB_flow.add(a10_database_tasks.CountLoadbalancersWithFlavor(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER),
            provides=a10constants.LB_COUNT_FLAVOR))
        delete_LB_flow.add(nat_pool_tasks.NatPoolDelete(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER,
                      a10constants.LB_COUNT_FLAVOR, constants.FLAVOR_DATA)))
        delete_LB_flow.add(a10_database_tasks.CountLBThunderPartition(
            requires=a10constants.VTHUNDER,
            provides=a10constants.LB_COUNT_THUNDER_PARTITION))
        delete_LB_flow.add(vthunder_tasks.DeleteHealthMonitorOnVThunder(
            requires=(a10constants.LB_COUNT_THUNDER_PARTITION, a10constants.VTHUNDER)))
        if CONF.a10_global.network_type == 'vlan':
            delete_LB_flow.add(
                vthunder_tasks.DeleteInterfaceTagIfNotInUseForLB(
                    requires=[
                        constants.LOADBALANCER,
                        a10constants.VTHUNDER]))
        delete_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            name=a10constants.MARK_VTHUNDER_MASTER_DELETED_IN_DB,
            requires=a10constants.VTHUNDER,
            inject={"status": constants.DELETED}))
        delete_LB_flow.add(database_tasks.MarkLBAmphoraeDeletedInDB(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.DisableLBAmphoraeHealthMonitoring(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.MarkLBDeletedInDB(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.DecrementLoadBalancerQuota(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        delete_LB_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            name=a10constants.SET_THUNDER_UPDATE_AT,
            requires=a10constants.VTHUNDER))
        if lb.topology == "ACTIVE_STANDBY":
            delete_LB_flow.add(a10_database_tasks.GetBackupVThunderByLoadBalancer(
                requires=constants.LOADBALANCER,
                provides=a10constants.BACKUP_VTHUNDER))
            delete_LB_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
                name=a10constants.MARK_VTHUNDER_BACKUP_DELETED_IN_DB,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
                inject={"status": constants.DELETED}))
            delete_LB_flow.add(a10_database_tasks.SetThunderUpdatedAt(
                name=a10constants.SET_THUNDER_BACKUP_UPDATE_AT,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}))
        return (delete_LB_flow, store)

    def get_post_lb_rack_vthunder_association_flow(self, prefix, topology,
                                                   mark_active=True):
        """Flow to manage networking after rack lb creation"""

        sf_name = prefix + '-' + constants.POST_LB_AMP_ASSOCIATION_SUBFLOW
        post_create_lb_flow = linear_flow.Flow(sf_name)
        post_create_lb_flow.add(vthunder_tasks.SetupDeviceNetworkMap(
            requires=a10constants.VTHUNDER,
            provides=a10constants.VTHUNDER))
        post_create_lb_flow.add(
            database_tasks.ReloadLoadBalancer(
                name=sf_name + '-' + constants.RELOAD_LB_AFTER_AMP_ASSOC,
                requires=constants.LOADBALANCER_ID,
                provides=constants.LOADBALANCER))
        post_create_lb_flow.add(database_tasks.UpdateLoadbalancerInDB(
            requires=[constants.LOADBALANCER, constants.UPDATE_DICT]))
        post_create_lb_flow.add(
            vthunder_tasks.CreateHealthMonitorOnVThunder(
                name=a10constants.CREATE_HEALTH_MONITOR_ON_VTHUNDER,
                requires=a10constants.VTHUNDER))
        if CONF.a10_global.network_type == 'vlan':
            post_create_lb_flow.add(vthunder_tasks.TagInterfaceForLB(
                requires=[constants.LOADBALANCER,
                          a10constants.VTHUNDER]))
        if CONF.a10_global.handle_vrid:
            post_create_lb_flow.add(self.handle_vrid_for_loadbalancer_subflow())
        if mark_active:
            post_create_lb_flow.add(database_tasks.MarkLBActiveInDB(
                name=sf_name + '-' + constants.MARK_LB_ACTIVE_INDB,
                requires=constants.LOADBALANCER))
        return post_create_lb_flow

    def handle_vrid_for_loadbalancer_subflow(self):
        handle_vrid_for_lb_subflow = linear_flow.Flow(
            a10constants.HANDLE_VRID_LOADBALANCER_SUBFLOW)
        handle_vrid_for_lb_subflow.add(a10_network_tasks.GetLBResourceSubnet(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.SUBNET))
        handle_vrid_for_lb_subflow.add(
            a10_database_tasks.GetChildProjectsOfParentPartition(
                requires=[a10constants.VTHUNDER],
                rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
                provides=a10constants.PARTITION_PROJECT_LIST
            ))
        handle_vrid_for_lb_subflow.add(
            a10_database_tasks.GetVRIDForLoadbalancerResource(
                requires=[
                    a10constants.PARTITION_PROJECT_LIST,
                    a10constants.VTHUNDER,
                    constants.LOADBALANCER],
                provides=a10constants.VRID_LIST))
        handle_vrid_for_lb_subflow.add(vthunder_tasks.ConfigureVRID(
            requires=a10constants.VTHUNDER))
        handle_vrid_for_lb_subflow.add(
            a10_network_tasks.HandleVRIDFloatingIP(
                requires=[
                    a10constants.VTHUNDER,
                    a10constants.VRID_LIST,
                    constants.SUBNET,
                    a10constants.VTHUNDER_CONFIG,
                    a10constants.USE_DEVICE_FLAVOR],
                rebind={
                    a10constants.LB_RESOURCE: constants.LOADBALANCER},
                provides=a10constants.VRID_LIST))
        handle_vrid_for_lb_subflow.add(
            a10_database_tasks.UpdateVRIDForLoadbalancerResource(
                requires=[
                    a10constants.VRID_LIST,
                    a10constants.VTHUNDER],
                rebind={
                    a10constants.LB_RESOURCE: constants.LOADBALANCER}))
        return handle_vrid_for_lb_subflow

    def get_delete_rack_lb_vrid_subflow(self):
        delete_lb_vrid_subflow = linear_flow.Flow(
            a10constants.DELETE_LOADBALANCER_VRID_SUBFLOW)
        delete_lb_vrid_subflow.add(
            a10_database_tasks.CountMembersInProjectBySubnet(
                requires=[constants.SUBNET, a10constants.PARTITION_PROJECT_LIST],
                provides=a10constants.MEMBER_COUNT))
        delete_lb_vrid_subflow.add(
            a10_network_tasks.GetMembersOnThunder(
                requires=[a10constants.VTHUNDER, a10constants.USE_DEVICE_FLAVOR],
                provides=a10constants.MEMBERS))
        delete_lb_vrid_subflow.add(
            a10_database_tasks.CountMembersOnThunderBySubnet(
                requires=[
                    constants.SUBNET,
                    a10constants.USE_DEVICE_FLAVOR,
                    a10constants.MEMBERS],
                provides=a10constants.MEMBER_COUNT_THUNDER))
        delete_lb_vrid_subflow.add(
            a10_database_tasks.GetVRIDForLoadbalancerResource(
                requires=[
                    a10constants.PARTITION_PROJECT_LIST,
                    a10constants.VTHUNDER,
                    constants.LOADBALANCER],
                provides=a10constants.VRID_LIST))
        delete_lb_vrid_subflow.add(
            a10_network_tasks.DeleteVRIDPort(
                requires=[
                    a10constants.VTHUNDER,
                    a10constants.VRID_LIST,
                    constants.SUBNET,
                    a10constants.USE_DEVICE_FLAVOR,
                    a10constants.LB_COUNT_SUBNET,
                    a10constants.MEMBER_COUNT,
                    a10constants.LB_COUNT_THUNDER,
                    a10constants.MEMBER_COUNT_THUNDER],
                rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
                provides=(
                    a10constants.VRID,
                    a10constants.DELETE_VRID)))
        delete_lb_vrid_subflow.add(a10_database_tasks.DeleteVRIDEntry(
            requires=[a10constants.VRID, a10constants.DELETE_VRID]))

        return delete_lb_vrid_subflow

    def get_delete_lb_vrid_subflow(self, deleteCompute):
        topology = CONF.a10_controller_worker.loadbalancer_topology

        delete_lb_vrid_subflow = linear_flow.Flow(
            a10constants.DELETE_LOADBALANCER_VRID_SUBFLOW)
        delete_lb_vrid_subflow.add(
            a10_database_tasks.CountMembersInProjectBySubnet(
                requires=[constants.SUBNET, a10constants.PARTITION_PROJECT_LIST],
                provides=a10constants.MEMBER_COUNT))
        delete_lb_vrid_subflow.add(
            a10_database_tasks.GetVRIDForLoadbalancerResource(
                requires=[
                    a10constants.PARTITION_PROJECT_LIST,
                    a10constants.VTHUNDER,
                    constants.LOADBALANCER],
                provides=a10constants.VRID_LIST))
        delete_lb_vrid_subflow.add(
            a10_network_tasks.DeleteVRIDPort(
                requires=[
                    a10constants.VTHUNDER,
                    a10constants.VRID_LIST,
                    constants.SUBNET,
                    a10constants.USE_DEVICE_FLAVOR,
                    a10constants.LB_COUNT_SUBNET,
                    a10constants.MEMBER_COUNT,
                    a10constants.LB_COUNT_THUNDER,
                    a10constants.MEMBER_COUNT_THUNDER],
                rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
                provides=(
                    a10constants.VRID,
                    a10constants.DELETE_VRID)))
        delete_lb_vrid_subflow.add(a10_database_tasks.DeleteVRIDEntry(
            requires=[a10constants.VRID, a10constants.DELETE_VRID]))

        if deleteCompute and topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            delete_lb_vrid_subflow.add(a10_database_tasks.DeleteProjectSetIdDB(
                requires=constants.LOADBALANCER))

        return delete_lb_vrid_subflow

    def _get_cascade_delete_pools_listeners_flow(self, lb):
        """Sets up an internal delete flow
        Because task flow doesn't support loops we store each pool
        and listener we want to delete in the store part and then rebind
        :param lb: load balancer
        :return: (flow, store) -- flow for the deletion and store with all
                  the listeners/pools stored properly
        """
        pools_listeners_delete_flow = linear_flow.Flow('pool_listener_delete_flow')
        store = {}
        # loop for loadbalancer's l7policy deletion
        l7policy_delete_flow = None
        for listener in lb.listeners:
            l7policy_delete_flow = linear_flow.Flow('l7policy_delete_flow')
            for l7policy in listener.l7policies:
                l7policy_name = 'l7policy_' + l7policy.id
                store[l7policy_name] = l7policy
                l7policy_delete_flow.add(
                    self._l7policy_flows.get_cascade_delete_l7policy_internal_flow(l7policy_name))
        if l7policy_delete_flow:
            pools_listeners_delete_flow.add(l7policy_delete_flow)

        # loop for loadbalancer's pool deletion
        for pool in lb.pools:
            pool_name = 'pool' + pool.id
            members = pool.members
            store[pool_name] = pool
            listeners = pool.listeners
            default_listener = None
            pool_listener_name = 'pool_listener' + pool.id
            if listeners:
                default_listener = pool.listeners[0]
            store[pool_listener_name] = default_listener
            health_mon = None
            health_monitor = pool.health_monitor
            if health_monitor is not None:
                health_mon = 'health_mon' + health_monitor.id
                store[health_mon] = health_monitor
            (pool_delete, pool_store) = self._pool_flows.get_cascade_delete_pool_internal_flow(
                pool_name, members, pool_listener_name, health_mon)
            store.update(pool_store)
            pools_listeners_delete_flow.add(pool_delete)

        # loop for loadbalancer's listener deletion
        listeners_delete_flow = unordered_flow.Flow('listener_delete_flow')
        compute_flag = lb.amphorae
        for listener in lb.listeners:
            listener_name = 'listener_' + listener.id
            store[listener_name] = listener
            listeners_delete_flow.add(
                self._listener_flows.get_cascade_delete_listener_internal_flow(
                    listener_name, compute_flag))
        pools_listeners_delete_flow.add(listeners_delete_flow)
        return (pools_listeners_delete_flow, store)



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
import imp

try:
    from unittest import mock
    from unittest.mock import patch
except ImportError:
    import mock
    from mock import patch

from octavia.certificates.common.auth.barbican_acl import BarbicanACLAuth
from octavia.common import data_models as o_data_models
from octavia.tests.common import constants as t_constants

from a10_octavia.common.data_models import Certificate
from a10_octavia.common.data_models import VThunder
from a10_octavia.controller.worker.tasks import cert_tasks
from a10_octavia.tests.common import a10constants as a10_test_constants
from a10_octavia.tests.unit.base import BaseTaskTestCase

LB = o_data_models.LoadBalancer(id=a10_test_constants.MOCK_LOAD_BALANCER_ID,
                                project_id=t_constants.MOCK_PROJECT_ID)
LISTENER = o_data_models.Listener(id=a10_test_constants.MOCK_LISTENER_ID,
                                  protocol='TERMINATED_HTTPS', protocol_port=2222,
                                  load_balancer=LB, tls_certificate_id='certificate-id-1')
VTHUNDER = VThunder()

CERT_DATA = Certificate(cert_filename=a10_test_constants.MOCK_CERT_FILENAME,
                        cert_content=a10_test_constants.MOCK_CERT_CONTENT,
                        key_filename=a10_test_constants.MOCK_KEY_FILENAME,
                        key_content=a10_test_constants.MOCK_KEY_CONTENT,
                        template_name=a10_test_constants.MOCK_TEMPLATE_NAME,
                        key_pass=a10_test_constants.MOCK_KEY_PASS)


class TestCertHandlerTasks(BaseTaskTestCase):

    def setUp(self):
        super(TestCertHandlerTasks, self).setUp()
        imp.reload(cert_tasks)
        self.listener = copy.deepcopy(LISTENER)
        self.client_mock = mock.Mock()

    @patch.object(BarbicanACLAuth, 'get_barbican_client')
    @mock.patch('a10_octavia.controller.worker.tasks.utils.get_cert_data', return_value=CERT_DATA)
    def test_GetSSLCertData_success(self, barbican_class, cert_data):
        mock_ssl_cert = cert_tasks.GetSSLCertData()
        out = mock_ssl_cert.execute(LB, LISTENER)
        self.assertEqual(out, CERT_DATA)

    @patch.object(BarbicanACLAuth, 'get_barbican_client')
    def test_GetSSLCertData_barbican_exception(self, barbican_class):
        mock_ssl_cert = cert_tasks.SSLCertCreate()
        mock_ssl_cert.axapi_client = self.client_mock
        barbican_class.side_effect = Exception
        self.assertRaises(Exception, lambda: mock_ssl_cert.execute(LB, LISTENER))
        self.client_mock.file.ssl_cert.exists.assert_not_called()

    # SSL Cert
    def test_ssl_cert_create_update(self):
        mock_ssl_cert = cert_tasks.SSLCertCreate()
        mock_ssl_cert.axapi_client = self.client_mock
        mock_ssl_cert.execute(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_cert.exists.assert_called_with(
            file=a10_test_constants.MOCK_CERT_FILENAME)
        self.client_mock.file.ssl_cert.update.assert_called_with(
            file=a10_test_constants.MOCK_CERT_FILENAME,
            cert=a10_test_constants.MOCK_CERT_CONTENT,
            size=len(a10_test_constants.MOCK_CERT_CONTENT),
            action=a10_test_constants.MOCK_CERT_ACTION,
            certificate_type=a10_test_constants.MOCK_CERT_TYPE)
        self.client_mock.file.ssl_cert.create.assert_not_called()

    def test_ssl_cert_create_new(self):
        mock_ssl_cert = cert_tasks.SSLCertCreate()
        mock_ssl_cert.axapi_client = self.client_mock
        self.client_mock.file.ssl_cert.exists.return_value = False
        mock_ssl_cert.execute(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_cert.create.assert_called_with(
            file=a10_test_constants.MOCK_CERT_FILENAME,
            cert=a10_test_constants.MOCK_CERT_CONTENT,
            size=len(a10_test_constants.MOCK_CERT_CONTENT),
            action=a10_test_constants.MOCK_CERT_ACTION,
            certificate_type=a10_test_constants.MOCK_CERT_TYPE)
        self.client_mock.file.ssl_cert.update.assert_not_called()

    def test_ssl_cert_create_revert(self):
        mock_ssl_cert = cert_tasks.SSLCertCreate()
        mock_ssl_cert.axapi_client = self.client_mock
        mock_ssl_cert.revert(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_cert.delete.assert_called_with(
            private_key=a10_test_constants.MOCK_KEY_FILENAME,
            cert_name=a10_test_constants.MOCK_CERT_FILENAME)

    def test_ssl_cert_update(self):
        mock_ssl_cert = cert_tasks.SSLCertUpdate()
        mock_ssl_cert.axapi_client = self.client_mock
        mock_ssl_cert.execute(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_cert.exists.assert_called_with(
            file=a10_test_constants.MOCK_CERT_FILENAME)
        self.client_mock.file.ssl_cert.update.assert_called_with(
            file=a10_test_constants.MOCK_CERT_FILENAME,
            cert=a10_test_constants.MOCK_CERT_CONTENT,
            size=len(a10_test_constants.MOCK_CERT_CONTENT),
            action=a10_test_constants.MOCK_CERT_ACTION,
            certificate_type=a10_test_constants.MOCK_CERT_TYPE)

    def test_ssl_cert_delete(self):
        mock_ssl_cert = cert_tasks.SSLCertDelete()
        mock_ssl_cert.axapi_client = self.client_mock
        mock_ssl_cert.execute(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_cert.exists.assert_called_with(
            file=a10_test_constants.MOCK_CERT_FILENAME)
        self.client_mock.file.ssl_cert.delete.assert_called_with(
            private_key=a10_test_constants.MOCK_KEY_FILENAME,
            cert_name=a10_test_constants.MOCK_CERT_FILENAME)

    # SSL Key
    def test_ssl_key_create_update(self):
        mock_ssl_key = cert_tasks.SSLKeyCreate()
        mock_ssl_key.axapi_client = self.client_mock
        mock_ssl_key.execute(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_key.exists.assert_called_with(
            file=a10_test_constants.MOCK_KEY_FILENAME)
        self.client_mock.file.ssl_key.update.assert_called_with(
            file=a10_test_constants.MOCK_KEY_FILENAME,
            cert=a10_test_constants.MOCK_KEY_CONTENT,
            size=len(a10_test_constants.MOCK_KEY_CONTENT),
            action=a10_test_constants.MOCK_CERT_ACTION)
        self.client_mock.file.ssl_key.create.assert_not_called()

    def test_ssl_key_create_new(self):
        mock_ssl_key = cert_tasks.SSLKeyCreate()
        mock_ssl_key.axapi_client = self.client_mock
        self.client_mock.file.ssl_key.exists.return_value = False
        mock_ssl_key.execute(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_key.create.assert_called_with(
            file=a10_test_constants.MOCK_KEY_FILENAME,
            cert=a10_test_constants.MOCK_KEY_CONTENT,
            size=len(a10_test_constants.MOCK_KEY_CONTENT),
            action=a10_test_constants.MOCK_CERT_ACTION)
        self.client_mock.file.ssl_key.update.assert_not_called()

    def test_ssl_key_create_revert(self):
        mock_ssl_key = cert_tasks.SSLKeyCreate()
        mock_ssl_key.axapi_client = self.client_mock
        mock_ssl_key.revert(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_key.delete.assert_called_with(
            private_key=a10_test_constants.MOCK_KEY_FILENAME)

    def test_ssl_key_update(self):
        mock_ssl_key = cert_tasks.SSLKeyUpdate()
        mock_ssl_key.axapi_client = self.client_mock
        mock_ssl_key.execute(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_key.exists.assert_called_with(
            file=a10_test_constants.MOCK_KEY_FILENAME)
        self.client_mock.file.ssl_key.update.assert_called_with(
            file=a10_test_constants.MOCK_KEY_FILENAME,
            cert=a10_test_constants.MOCK_KEY_CONTENT,
            size=len(a10_test_constants.MOCK_KEY_CONTENT),
            action=a10_test_constants.MOCK_CERT_ACTION)

    def test_ssl_key_delete(self):
        mock_ssl_key = cert_tasks.SSLKeyDelete()
        mock_ssl_key.axapi_client = self.client_mock
        mock_ssl_key.execute(CERT_DATA, VTHUNDER)
        self.client_mock.file.ssl_key.delete.assert_called_with(
            private_key=a10_test_constants.MOCK_KEY_FILENAME)

    # Client Template
    def test_client_ssl_template_create_update(self):
        mock_client_ssl_template = cert_tasks.ClientSSLTemplateCreate()
        mock_client_ssl_template.axapi_client = self.client_mock
        mock_client_ssl_template.execute(CERT_DATA, VTHUNDER)
        self.client_mock.slb.template.client_ssl.exists.assert_called_with(
            name=a10_test_constants.MOCK_LISTENER_ID)
        self.client_mock.slb.template.client_ssl.update.assert_called_with(
            name=a10_test_constants.MOCK_LISTENER_ID,
            cert=a10_test_constants.MOCK_CERT_FILENAME,
            key=a10_test_constants.MOCK_KEY_FILENAME,
            passphrase=a10_test_constants.MOCK_KEY_PASS)
        self.client_mock.slb.template.client_ssl.create.assert_not_called()

    def test_client_ssl_template_create_new(self):
        mock_client_ssl_template = cert_tasks.ClientSSLTemplateCreate()
        mock_client_ssl_template.axapi_client = self.client_mock
        self.client_mock.slb.template.client_ssl.exists.return_value = False
        mock_client_ssl_template.execute(CERT_DATA, VTHUNDER)
        self.client_mock.slb.template.client_ssl.create.assert_called_with(
            name=a10_test_constants.MOCK_LISTENER_ID,
            cert=a10_test_constants.MOCK_CERT_FILENAME,
            key=a10_test_constants.MOCK_KEY_FILENAME,
            passphrase=a10_test_constants.MOCK_KEY_PASS)
        self.client_mock.slb.template.client_ssl.update.assert_not_called()

    def test_client_ssl_template_create_revert(self):
        mock_client_ssl_template = cert_tasks.ClientSSLTemplateCreate()
        mock_client_ssl_template.axapi_client = self.client_mock
        mock_client_ssl_template.revert(CERT_DATA, VTHUNDER)
        self.client_mock.slb.template.client_ssl.delete.assert_called_with(
            name=a10_test_constants.MOCK_LISTENER_ID)

    def test_client_ssl_template_update(self):
        mock_client_ssl_template = cert_tasks.ClientSSLTemplateUpdate()
        mock_client_ssl_template.axapi_client = self.client_mock
        mock_client_ssl_template.execute(CERT_DATA, VTHUNDER)
        self.client_mock.slb.template.client_ssl.exists.assert_called_with(
            name=a10_test_constants.MOCK_LISTENER_ID)
        self.client_mock.slb.template.client_ssl.update.assert_called_with(
            name=a10_test_constants.MOCK_LISTENER_ID,
            cert=a10_test_constants.MOCK_CERT_FILENAME,
            key=a10_test_constants.MOCK_KEY_FILENAME,
            passphrase=a10_test_constants.MOCK_KEY_PASS)

    def test_client_ssl_template_delete(self):
        mock_client_ssl_template = cert_tasks.ClientSSLTemplateDelete()
        mock_client_ssl_template.axapi_client = self.client_mock
        mock_client_ssl_template.execute(CERT_DATA, VTHUNDER)
        self.client_mock.slb.template.client_ssl.delete.assert_called_with(
            name=a10_test_constants.MOCK_LISTENER_ID)



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from taskflow.patterns import linear_flow

from octavia.common import constants
from octavia.controller.worker.v1.tasks import database_tasks
from octavia.controller.worker.v1.tasks import lifecycle_tasks
from octavia.controller.worker.v1.tasks import model_tasks

from a10_octavia.common import a10constants
from a10_octavia.controller.worker.tasks import a10_database_tasks
from a10_octavia.controller.worker.tasks import l7policy_tasks
from a10_octavia.controller.worker.tasks import vthunder_tasks


class L7PolicyFlows(object):

    def get_create_l7policy_flow(self, topology):
        """Create a flow to create an L7 policy

        :returns: The flow for creating an L7 policy
        """
        create_l7policy_flow = linear_flow.Flow(constants.CREATE_L7POLICY_FLOW)
        create_l7policy_flow.add(lifecycle_tasks.L7PolicyToErrorOnRevertTask(
            requires=[constants.L7POLICY,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        create_l7policy_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))

        create_l7policy_flow.add(database_tasks.MarkL7PolicyPendingCreateInDB(
            requires=constants.L7POLICY))
        create_l7policy_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            create_l7policy_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        create_l7policy_flow.add(l7policy_tasks.CreateL7Policy(
            requires=[constants.L7POLICY, constants.LISTENERS, a10constants.VTHUNDER]))
        create_l7policy_flow.add(database_tasks.MarkL7PolicyActiveInDB(
            requires=constants.L7POLICY))
        create_l7policy_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=[constants.LOADBALANCER, constants.LISTENERS]))
        create_l7policy_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        create_l7policy_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))
        return create_l7policy_flow

    def get_delete_l7policy_flow(self, topology):
        """Create a flow to delete an L7 policy

        :returns: The flow for deleting an L7 policy
        """
        delete_l7policy_flow = linear_flow.Flow(constants.DELETE_L7POLICY_FLOW)
        delete_l7policy_flow.add(lifecycle_tasks.L7PolicyToErrorOnRevertTask(
            requires=[constants.L7POLICY,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        delete_l7policy_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))

        delete_l7policy_flow.add(database_tasks.MarkL7PolicyPendingDeleteInDB(
            requires=constants.L7POLICY))
        delete_l7policy_flow.add(model_tasks.DeleteModelObject(
            rebind={constants.OBJECT: constants.L7POLICY}))
        delete_l7policy_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            delete_l7policy_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        delete_l7policy_flow.add(l7policy_tasks.DeleteL7Policy(
            requires=[constants.L7POLICY, a10constants.VTHUNDER]))
        delete_l7policy_flow.add(database_tasks.DeleteL7PolicyInDB(
            requires=constants.L7POLICY))
        delete_l7policy_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=[constants.LOADBALANCER, constants.LISTENERS]))
        delete_l7policy_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        delete_l7policy_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))
        return delete_l7policy_flow

    def get_update_l7policy_flow(self, topology):
        """Create a flow to update an L7 policy
        :returns: The flow for updating an L7 policy
        """
        update_l7policy_flow = linear_flow.Flow(constants.UPDATE_L7POLICY_FLOW)
        update_l7policy_flow.add(lifecycle_tasks.L7PolicyToErrorOnRevertTask(
            requires=[constants.L7POLICY,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        update_l7policy_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))

        update_l7policy_flow.add(database_tasks.MarkL7PolicyPendingUpdateInDB(
            requires=constants.L7POLICY))
        update_l7policy_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            update_l7policy_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        update_l7policy_flow.add(
            l7policy_tasks.UpdateL7Policy(
                requires=[
                    constants.L7POLICY,
                    constants.LISTENERS,
                    a10constants.VTHUNDER,
                    constants.UPDATE_DICT]))

        update_l7policy_flow.add(database_tasks.UpdateL7PolicyInDB(
            requires=[constants.L7POLICY, constants.UPDATE_DICT]))
        update_l7policy_flow.add(database_tasks.MarkL7PolicyActiveInDB(
            requires=constants.L7POLICY))
        update_l7policy_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=[constants.LOADBALANCER, constants.LISTENERS]))
        update_l7policy_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        update_l7policy_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))
        return update_l7policy_flow

    def get_cascade_delete_l7policy_internal_flow(self, l7policy_name):
        delete_l7policy_flow = linear_flow.Flow(constants.DELETE_L7POLICY_FLOW)
        delete_l7policy_flow.add(model_tasks.DeleteModelObject(
            name='delete_model_object_' + l7policy_name,
            rebind={constants.OBJECT: l7policy_name}))
        delete_l7policy_flow.add(l7policy_tasks.DeleteL7Policy(
            name='delete_l7policy_' + l7policy_name,
            requires=[constants.L7POLICY, a10constants.VTHUNDER],
            rebind={constants.L7POLICY: l7policy_name}))
        return delete_l7policy_flow



from __future__ import with_statement

from logging.config import fileConfig

from sqlalchemy import engine_from_config
from sqlalchemy import pool

from alembic import context
from a10_octavia import a10_config
from a10_octavia.db import base_models

# ===========================================================
#                    DO NOT DELETE THIS CODE!!!
#
# The below code is required for autogeneration to work
# ===========================================================
from a10_octavia.db.models import VThunder, NATPool, VRID, VrrpSet
VALID_AUOTGEN_TABLE_NAMES = ('vthunders', 'nat_pool', 'vrid', 'vrrp_set')
# ===========================================================

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
VERSION_TABLE = 'alembic_version_a10'

config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = [base_models.BASE.metadata]

if getattr(config, 'connection', None) is None:
    a10_cfg = a10_config.A10Config()
    config.set_main_option("sqlalchemy.url", a10_cfg.get('database_connection'))


# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.

def include_object(object, name, type_, reflected, compare_to):
    if type_ == "table" and name not in VALID_AUOTGEN_TABLE_NAMES:
        return False
    return True

def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        include_object=include_object
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            version_table=VERSION_TABLE,
            target_metadata=target_metadata,
            include_object=include_object
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import partial
import multiprocessing
import os
import signal
import sys
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_reports import guru_meditation_report as gmr
import setproctitle
from stevedore import enabled as stevedore_enabled

from octavia.api.drivers.driver_agent import driver_listener
from octavia.common import service
from octavia import version

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
PROVIDER_AGENT_PROCESSES = []


def _mutate_config(*args, **kwargs):
    CONF.mutate_config_files()


def _handle_mutate_config(status_proc_pid, stats_proc_pid, *args, **kwargs):
    LOG.info("Driver agent received HUP signal, mutating config.")
    _mutate_config()
    os.kill(status_proc_pid, signal.SIGHUP)
    os.kill(stats_proc_pid, signal.SIGHUP)


def _check_if_provider_agent_enabled(extension):
    if extension.name in CONF.driver_agent.enabled_provider_agents:
        return True
    return False


def _process_wrapper(exit_event, proc_name, function, agent_name=None):
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGHUP, _mutate_config)
    if agent_name:
        process_title = 'octavia-driver-agent - {} -- {}'.format(
            proc_name, agent_name)
    else:
        process_title = 'octavia-driver-agent - {}'.format(proc_name)
    setproctitle.setproctitle(process_title)
    while not exit_event.is_set():
        try:
            function(exit_event)
        except Exception as e:
            if agent_name:
                LOG.exception('Provider agent "%s" raised exception: %s. '
                              'Restarting the "%s" provider agent.',
                              agent_name, str(e), agent_name)
            else:
                LOG.exception('%s raised exception: %s. '
                              'Restarting %s.',
                              proc_name, str(e), proc_name)
            time.sleep(1)
            continue
        break


def _start_provider_agents(exit_event):
    extensions = stevedore_enabled.EnabledExtensionManager(
        namespace='octavia.driver_agent.provider_agents',
        check_func=_check_if_provider_agent_enabled)
    for ext in extensions:
        ext_process = multiprocessing.Process(
            name=ext.name, target=_process_wrapper,
            args=(exit_event, 'provider_agent', ext.plugin),
            kwargs={'agent_name': ext.name})
        PROVIDER_AGENT_PROCESSES.append(ext_process)
        ext_process.start()
        LOG.info('Started enabled provider agent: "%s" with PID: %d.',
                 ext.name, ext_process.pid)


def main():
    service.prepare_service(sys.argv)

    gmr.TextGuruMeditation.setup_autorun(version)

    processes = []
    exit_event = multiprocessing.Event()

    status_listener_proc = multiprocessing.Process(
        name='status_listener', target=_process_wrapper,
        args=(exit_event, 'status_listener', driver_listener.status_listener))
    processes.append(status_listener_proc)

    LOG.info("Driver agent status listener process starts:")
    status_listener_proc.start()

    stats_listener_proc = multiprocessing.Process(
        name='stats_listener', target=_process_wrapper,
        args=(exit_event, 'stats_listener', driver_listener.stats_listener))
    processes.append(stats_listener_proc)

    LOG.info("Driver agent statistics listener process starts:")
    stats_listener_proc.start()

    get_listener_proc = multiprocessing.Process(
        name='get_listener', target=_process_wrapper,
        args=(exit_event, 'get_listener', driver_listener.get_listener))
    processes.append(get_listener_proc)

    LOG.info("Driver agent get listener process starts:")
    get_listener_proc.start()

    _start_provider_agents(exit_event)

    def process_cleanup(*args, **kwargs):
        LOG.info("Driver agent exiting due to signal.")
        exit_event.set()
        status_listener_proc.join()
        stats_listener_proc.join()
        get_listener_proc.join()

        for proc in PROVIDER_AGENT_PROCESSES:
            LOG.info('Waiting up to %s seconds for provider agent "%s" to '
                     'shutdown.',
                     CONF.driver_agent.provider_agent_shutdown_timeout,
                     proc.name)
            try:
                proc.join(CONF.driver_agent.provider_agent_shutdown_timeout)
                if proc.exitcode is None:
                    # TODO(johnsom) Change to proc.kill() once
                    #               python 3.7 or newer only
                    os.kill(proc.pid, signal.SIGKILL)
                    LOG.warning(
                        'Forcefully killed "%s" provider agent because it '
                        'failed to shutdown in %s seconds.', proc.name,
                        CONF.driver_agent.provider_agent_shutdown_timeout)
            except Exception as e:
                LOG.warning('Unknown error "%s" while shutting down "%s", '
                            'ignoring and continuing shutdown process.',
                            str(e), proc.name)
            else:
                LOG.info('Provider agent "%s" has succesfully shutdown.',
                         proc.name)

    signal.signal(signal.SIGTERM, process_cleanup)
    signal.signal(signal.SIGHUP, partial(
        _handle_mutate_config, status_listener_proc.pid,
        stats_listener_proc.pid, get_listener_proc.pid))

    try:
        for process in processes:
            process.join()
    except KeyboardInterrupt:
        process_cleanup()



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors
from requests.exceptions import ConnectionError
from taskflow import task

from a10_octavia.common.a10constants import PERS_TYPE
from a10_octavia.common.a10constants import SP_OBJ_DICT
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator_for_revert

LOG = logging.getLogger(__name__)


class HandleSessionPersistenceDelta(task.Task):
    """Task to handle session persistence template delta for pool"""

    @axapi_client_decorator
    def execute(self, vthunder, pool):
        sess_pers = pool.session_persistence
        if pool.session_persistence and hasattr(pool.session_persistence, 'to_dict'):
            sess_pers = pool.session_persistence.to_dict()

        # Remove existing persistence template if any
        for sp_type in PERS_TYPE:
            try:
                sp_template = getattr(self.axapi_client.slb.template, sp_type)
                sp_template.delete(pool.id)
                LOG.debug("Successfully deleted existing session persistence template "
                          "for pool: %s", pool.id)
            except acos_errors.NotFound:
                pass
            except (acos_errors.ACOSException, ConnectionError) as e:
                LOG.exception("Failed to delete existing session persistence for pool: %s",
                              pool.id)
                raise e

        if sess_pers and sess_pers['type'] in SP_OBJ_DICT:
            sp_template = getattr(self.axapi_client.slb.template, SP_OBJ_DICT[sess_pers['type']])

            try:
                if sess_pers['cookie_name']:
                    sp_template.create(pool.id, cookie_name=sess_pers['cookie_name'])
                else:
                    sp_template.create(pool.id)
                LOG.debug("Successfully created session persistence template for pool: %s", pool.id)
            except acos_errors.Exists:
                pass
            except (acos_errors.ACOSException, ConnectionError) as e:
                LOG.exception("Failed to create session persistence for pool: %s", pool.id)
                raise e

    @axapi_client_decorator_for_revert
    def revert(self, vthunder, pool, *args, **kwargs):
        LOG.warning("Reverting creation of session persistence for pool: %s", pool.id)
        try:
            sp_template = getattr(self.axapi_client.slb.template,
                                  SP_OBJ_DICT[pool.session_persistence.type])
            sp_template.delete(pool.id)
            LOG.debug("Successfully deleted session persistence template for pool: %s", pool.id)
        except acos_errors.NotFound:
            pass
        except ConnectionError:
            LOG.exception("Failed to connect A10 Thunder device: %s", vthunder.ip_address)
        except Exception as e:
            LOG.exception("Failed to revert creation of session persistence template for pool:"
                          " %s due to: %s", (pool.id, str(e)))


class DeleteSessionPersistence(task.Task):
    """Task to delete session persistence templates"""

    @axapi_client_decorator
    def execute(self, vthunder, pool):
        if pool.session_persistence:
            for sp_type in PERS_TYPE:
                try:
                    sp_template = getattr(self.axapi_client.slb.template, sp_type)
                    sp_template.delete(pool.id)
                    LOG.debug("Successfully deleted session persistence template for pool: %s",
                              pool.id)
                except acos_errors.NotFound:
                    pass
                except (acos_errors.ACOSException, ConnectionError) as e:
                    LOG.warning("Failed to delete session persistence: %s", str(e))
                    raise e



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from requests.exceptions import ConnectionError
from taskflow import task

from octavia.common import constants
from octavia.common import exceptions

import acos_client.errors as acos_errors

from a10_octavia.common import openstack_mappings
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator_for_revert
from a10_octavia.controller.worker.tasks import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class PoolParent(object):

    def set(self, set_method, pool, vthunder, flavor=None, **kwargs):
        pool_args = {'service_group': utils.meta(pool, 'service_group', {})}

        device_templates = self.axapi_client.slb.template.templates.get()

        service_group_temp = {}
        template_server = CONF.service_group.template_server
        if template_server and template_server.lower() != 'none':
            if CONF.a10_global.use_shared_for_template_lookup:
                LOG.warning('Shared partition template lookup for `[service_group]`'
                            ' is not supported on template `template-server`')
            service_group_temp['template-server'] = template_server

        template_port = CONF.service_group.template_port
        if template_port and template_port.lower() != 'none':
            if CONF.a10_global.use_shared_for_template_lookup:
                LOG.warning('Shared partition template lookup for `[service_group]`'
                            ' is not supported on template `template-port`')
            service_group_temp['template-port'] = template_port

        template_policy = CONF.service_group.template_policy
        if template_policy and template_policy.lower() != 'none':
            template_key = 'template-policy'
            if vthunder.partition_name != "shared":
                if CONF.a10_global.use_shared_for_template_lookup:
                    template_key = utils.shared_template_modifier(template_key,
                                                                  template_policy,
                                                                  device_templates)
            service_group_temp[template_key] = template_policy

        protocol = openstack_mappings.service_group_protocol(
            self.axapi_client, pool.protocol)
        lb_method = openstack_mappings.service_group_lb_method(
            self.axapi_client, pool.lb_algorithm)

        # Handle options from flavor
        if flavor:
            pool_flavor = flavor.get('service_group')
            if pool_flavor:
                name_exprs = pool_flavor.get('name_expressions')
                parsed_exprs = utils.parse_name_expressions(pool.name, name_exprs)
                pool_flavor.pop('name_expressions', None)
                pool_args['service_group'].update(pool_flavor)
                pool_args['service_group'].update(parsed_exprs)

        set_method(pool.id,
                   protocol=protocol,
                   lb_method=lb_method,
                   service_group_templates=service_group_temp,
                   mem_list=kwargs.get('mem_list'),
                   hm_name=kwargs.get('health_monitor'),
                   **pool_args)


class PoolCreate(PoolParent, task.Task):
    """Task to create pool"""

    @axapi_client_decorator
    def execute(self, pool, vthunder, flavor=None):
        try:
            if pool.protocol == constants.PROTOCOL_PROXY:
                raise exceptions.ProviderUnsupportedOptionError(
                    prov="A10",
                    user_msg=("A pool with protocol PROXY is not supported by A10 provider."
                              "Failed to create pool {0}").format(pool.id))

            self.set(self.axapi_client.slb.service_group.create, pool, vthunder, flavor)
            LOG.debug("Successfully created pool: %s", pool.id)
            return pool
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to create pool: %s", pool.id)
            raise e

    @axapi_client_decorator_for_revert
    def revert(self, pool, vthunder, flavor=None, *args, **kwargs):
        LOG.warning("Reverting creation of pool: %s", pool.id)
        try:
            self.axapi_client.slb.service_group.delete(pool.id)
        except ConnectionError:
            LOG.exception(
                "Failed to connect A10 Thunder device: %s", vthunder.ip_address)
        except Exception as e:
            LOG.exception("Failed to revert creation of pool: %s due to: %s",
                          pool.id, str(e))


class PoolDelete(task.Task):
    """Task to delete pool"""

    @axapi_client_decorator
    def execute(self, pool, vthunder):
        try:
            self.axapi_client.slb.service_group.delete(pool.id)
            LOG.debug("Successfully deleted pool: %s", pool.id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to delete pool: %s", pool.id)
            raise e


class PoolUpdate(PoolParent, task.Task):
    """Task to update pool"""

    @axapi_client_decorator
    def execute(self, pool, vthunder, update_dict={}, flavor=None):
        pool.__dict__.update(update_dict)
        try:
            service_group = self.axapi_client.slb.service_group.get(
                pool.id)['service-group']
            mem_list = service_group.get('member-list')
            health_monitor = service_group.get('health-check')
            self.set(self.axapi_client.slb.service_group.replace, pool, vthunder,
                     mem_list=mem_list, health_monitor=health_monitor, flavor=flavor)
            LOG.debug("Successfully updated pool: %s", pool.id)
        except (acos_errors.ACOSException, ConnectionError) as e:
            LOG.exception("Failed to update pool: %s", pool.id)
            raise e



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import logging
from oslo_config import cfg
from taskflow.patterns import graph_flow
from taskflow.patterns import linear_flow

from octavia.common import constants
from octavia.controller.worker.v1.tasks import compute_tasks as compute
from octavia.controller.worker.v1.tasks import database_tasks

from a10_octavia.common import a10constants
from a10_octavia.controller.worker.tasks import a10_compute_tasks as compute_tasks
from a10_octavia.controller.worker.tasks import a10_database_tasks
from a10_octavia.controller.worker.tasks import a10_network_tasks
from a10_octavia.controller.worker.tasks import glm_tasks
from a10_octavia.controller.worker.tasks import vthunder_tasks

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class VThunderFlows(object):

    def __init__(self):
        pass

    def get_create_vthunder_flow(self):
        """Flow to create a spare amphora."""

        create_vthunder_flow = linear_flow.Flow(
            constants.CREATE_AMPHORA_FLOW)
        sf_name = a10constants.SPARE_VTHUNDER_CREATE
        create_vthunder_flow.add(database_tasks.CreateAmphoraInDB(
            name=sf_name + '-' + constants.CREATE_AMPHORA_INDB,
            provides=constants.AMPHORA_ID))
        create_vthunder_flow.add(compute_tasks.ComputeCreate(
            name=sf_name + '-' + constants.COMPUTE_CREATE,
            requires=(constants.AMPHORA_ID,
                      constants.BUILD_TYPE_PRIORITY,
                      ),
            provides=constants.COMPUTE_ID))
        create_vthunder_flow.add(database_tasks.UpdateAmphoraComputeId(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_COMPUTEID,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_ID)))
        create_vthunder_flow.add(database_tasks.MarkAmphoraBootingInDB(
            name=sf_name + '-' + constants.MARK_AMPHORA_BOOTING_INDB,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_ID)))
        create_vthunder_flow.add(compute_tasks.ComputeActiveWait(
            name=sf_name + '-' + constants.COMPUTE_WAIT,
            requires=(constants.COMPUTE_ID, constants.AMPHORA_ID),
            provides=constants.COMPUTE_OBJ))
        create_vthunder_flow.add(database_tasks.UpdateAmphoraInfo(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_INFO,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_OBJ),
            provides=constants.AMPHORA))
        create_vthunder_flow.add(a10_database_tasks.CreateSpareVThunderEntry(
            requires=(constants.AMPHORA),
            provides=a10constants.VTHUNDER))
        create_vthunder_flow.add(
            vthunder_tasks.VThunderComputeConnectivityWait(
                name=sf_name + '-' + constants.AMP_COMPUTE_CONNECTIVITY_WAIT,
                requires=(a10constants.VTHUNDER, constants.AMPHORA)))
        create_vthunder_flow.add(vthunder_tasks.EnableInterfaceOnSpare(
            name=sf_name + '-' + a10constants.ENABLE_VTHUNDER_INTERFACE,
            requires=(a10constants.VTHUNDER)))
        create_vthunder_flow.add(
            vthunder_tasks.UpdateAcosVersionInVthunderEntry(
                name=sf_name + '-' + a10constants.UPDATE_ACOS_VERSION_IN_VTHUNDER_ENTRY,
                requires=(a10constants.VTHUNDER)))
        create_vthunder_flow.add(
            database_tasks.MarkAmphoraReadyInDB(
                name=sf_name + '-' + a10constants.MARK_AMPHORA_READY_INDB,
                requires=(constants.AMPHORA)))
        create_vthunder_flow.add(self.get_glm_license_subflow(
            prefix=a10constants.SPARE_VTHUNDER_CREATE, role=a10constants.SPARE_VTHUNDER))
        create_vthunder_flow.add(
            vthunder_tasks.CreateHealthMonitorOnVThunder(
                name=sf_name + '-' + a10constants.CREATE_HEALTH_MONITOR_ON_SPARE,
                requires=(a10constants.VTHUNDER)))

        create_vthunder_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            requires=(a10constants.VTHUNDER),
            inject={"status": a10constants.READY}))
        return create_vthunder_flow

    def get_vthunder_for_lb_subflow(
            self, prefix, role=constants.ROLE_STANDALONE):
        """Flow to get vThunder for lb"""

        sf_name = prefix + '-' + constants.GET_AMPHORA_FOR_LB_SUBFLOW

        # We need a graph flow here for a conditional flow
        amp_for_lb_flow = graph_flow.Flow(sf_name)

        amp_for_lb_flow.add(database_tasks.ReloadLoadBalancer(
            name=sf_name + '-' + 'reload_loadbalancer',
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER))

        # Setup the task that maps an amphora to a load balancer
        allocate_and_associate_amp = a10_database_tasks.MapLoadbalancerToAmphora(
            name=sf_name + '-' + constants.MAP_LOADBALANCER_TO_AMPHORA,
            requires=constants.LOADBALANCER,
            inject={a10constants.ROLE: role},
            provides=constants.AMPHORA_ID)

        create_amp = self._get_create_amp_for_lb_subflow(prefix, role)

        map_lb_to_vthunder = self._get_vthunder_for_amphora_subflow(
            prefix, role)

        # Add them to the graph flow
        amp_for_lb_flow.add(allocate_and_associate_amp,
                            map_lb_to_vthunder, create_amp)

        # Setup the decider for the path if we can map vThunder
        amp_for_lb_flow.link(allocate_and_associate_amp, map_lb_to_vthunder,
                             decider=self._allocate_amp_to_lb_decider,
                             decider_depth='flow')
        # Setup the decider to create a vThunder
        amp_for_lb_flow.link(allocate_and_associate_amp, create_amp,
                             decider=self._create_new_amp_for_lb_decider,
                             decider_depth='flow')

        return amp_for_lb_flow

    def _get_create_amp_for_lb_subflow(self, prefix, role):
        """Flow to create a new vThunder for lb."""

        sf_name = prefix + '-' + constants.CREATE_AMP_FOR_LB_SUBFLOW
        create_amp_for_lb_subflow = linear_flow.Flow(sf_name)
        create_amp_for_lb_subflow.add(database_tasks.CreateAmphoraInDB(
            name=sf_name + '-' + constants.CREATE_AMPHORA_INDB,
            provides=constants.AMPHORA_ID))
        # VIP subnet integration at bootup
        create_amp_for_lb_subflow.add(database_tasks.ReloadLoadBalancer(
            name=sf_name + '-' + a10constants.RELOADLOAD_BALANCER,
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER))
        require_server_group_id_condition = (
            role in (constants.ROLE_BACKUP, constants.ROLE_MASTER) and
            CONF.a10_nova.enable_anti_affinity)

        if require_server_group_id_condition:
            create_amp_for_lb_subflow.add(compute_tasks.ComputeCreate(
                name=sf_name + '-' + constants.COMPUTE_CREATE,
                requires=(
                    constants.AMPHORA_ID,
                    constants.BUILD_TYPE_PRIORITY,
                    constants.SERVER_GROUP_ID,
                    constants.LOADBALANCER
                ),
                provides=constants.COMPUTE_ID))
        else:
            create_amp_for_lb_subflow.add(compute_tasks.ComputeCreate(
                name=sf_name + '-' + constants.COMPUTE_CREATE,
                requires=(
                    constants.AMPHORA_ID,
                    constants.BUILD_TYPE_PRIORITY,
                    constants.LOADBALANCER
                ),
                provides=constants.COMPUTE_ID))

        create_amp_for_lb_subflow.add(database_tasks.UpdateAmphoraComputeId(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_COMPUTEID,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_ID)))
        create_amp_for_lb_subflow.add(database_tasks.MarkAmphoraBootingInDB(
            name=sf_name + '-' + constants.MARK_AMPHORA_BOOTING_INDB,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_ID)))
        create_amp_for_lb_subflow.add(compute_tasks.ComputeActiveWait(
            name=sf_name + '-' + constants.COMPUTE_WAIT,
            requires=(constants.COMPUTE_ID, constants.AMPHORA_ID),
            provides=constants.COMPUTE_OBJ))
        create_amp_for_lb_subflow.add(database_tasks.UpdateAmphoraInfo(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_INFO,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_OBJ),
            provides=constants.AMPHORA))
        # Create vThunder entry in custom DB
        create_amp_for_lb_subflow.add(a10_database_tasks.CreateVThunderEntry(
            name=sf_name + '-' + a10constants.CREATE_VTHUNDER_ENTRY,
            requires=(constants.AMPHORA, constants.LOADBALANCER),
            inject={a10constants.ROLE: role, a10constants.STATUS: constants.PENDING_CREATE}))
        # Rebind requires vthunder in store and vMaster requires vThunder
        create_amp_for_lb_subflow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            name=sf_name + '-' + a10constants.VTHUNDER_BY_LB,
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        # Get VThunder details from database
        if role == constants.ROLE_BACKUP:
            create_amp_for_lb_subflow.add(
                a10_database_tasks.GetBackupVThunderByLoadBalancer(
                    requires=constants.LOADBALANCER,
                    provides=a10constants.BACKUP_VTHUNDER))
            create_amp_for_lb_subflow.add(
                vthunder_tasks.VThunderComputeConnectivityWait(
                    name=sf_name + '-' + a10constants.BACKUP_CONNECTIVITY_WAIT,
                    rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
                    requires=constants.AMPHORA))
        else:
            create_amp_for_lb_subflow.add(
                vthunder_tasks.VThunderComputeConnectivityWait(
                    name=sf_name + '-' + a10constants.WAIT_FOR_VTHUNDER_CONNECTIVITY,
                    requires=(a10constants.VTHUNDER, constants.AMPHORA)))
        # License the vThunder-Amphora
        create_amp_for_lb_subflow.add(
            *self.get_glm_license_subflow(prefix + '-' + role, role))
        create_amp_for_lb_subflow.add(
            database_tasks.MarkAmphoraAllocatedInDB(
                name=sf_name + '-' + constants.MARK_AMPHORA_ALLOCATED_INDB,
                requires=(constants.AMPHORA, constants.LOADBALANCER_ID)))
        create_amp_for_lb_subflow.add(database_tasks.ReloadAmphora(
            name=sf_name + '-' + constants.RELOAD_AMPHORA,
            requires=constants.AMPHORA_ID,
            provides=constants.AMPHORA))
        if role == constants.ROLE_MASTER:
            create_amp_for_lb_subflow.add(database_tasks.MarkAmphoraMasterInDB(
                name=sf_name + '-' + constants.MARK_AMP_MASTER_INDB,
                requires=constants.AMPHORA))
        elif role == constants.ROLE_BACKUP:
            create_amp_for_lb_subflow.add(database_tasks.MarkAmphoraBackupInDB(
                name=sf_name + '-' + constants.MARK_AMP_BACKUP_INDB,
                requires=constants.AMPHORA))
        elif role == constants.ROLE_STANDALONE:
            create_amp_for_lb_subflow.add(
                database_tasks.MarkAmphoraStandAloneInDB(
                    name=sf_name + '-' + constants.MARK_AMP_STANDALONE_INDB,
                    requires=constants.AMPHORA))

        if role == constants.ROLE_BACKUP:
            create_amp_for_lb_subflow.add(
                a10_database_tasks.GetBackupVThunderByLoadBalancer(
                    name=sf_name + '-' + a10constants.BACKUP_VTHUNDER,
                    requires=constants.LOADBALANCER,
                    provides=a10constants.BACKUP_VTHUNDER))
            create_amp_for_lb_subflow.add(
                vthunder_tasks.UpdateAcosVersionInVthunderEntry(
                    name=sf_name + '-' + a10constants.UPDATE_ACOS_VERSION_FOR_BACKUP_VTHUNDER,
                    rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}))
        else:
            create_amp_for_lb_subflow.add(
                vthunder_tasks.UpdateAcosVersionInVthunderEntry(
                    name=sf_name + '-' + a10constants.UPDATE_ACOS_VERSION_FOR_BACKUP_VTHUNDER,
                    requires=(a10constants.VTHUNDER)))

        return create_amp_for_lb_subflow

    def _allocate_amp_to_lb_decider(self, history):
        """Decider to allocate lb to vThunder"""

        return list(history.values())[0] is not None

    def _create_new_amp_for_lb_decider(self, history):
        """Decider to create vThunder for lb"""

        return list(history.values())[0] is None

    def _get_vthunder_for_amphora_subflow(self, prefix, role):
        """Subflow to create lb in existing vThunder."""

        sf_name = prefix + '-' + a10constants.LB_TO_VTHUNDER_SUBFLOW
        vthunder_for_amphora_subflow = linear_flow.Flow(sf_name)
        vthunder_for_amphora_subflow.add(database_tasks.CreateAmphoraInDB(
            name=sf_name + '-' + constants.CREATE_AMPHORA_INDB,
            provides=constants.AMPHORA_ID))
        vthunder_for_amphora_subflow.add(a10_database_tasks.ValidateComputeForProject(
            name=sf_name + '-' + a10constants.VALIDATE_COMPUTE_FOR_PROJECT,
            requires=constants.LOADBALANCER,
            inject={"role": role},
            provides=constants.COMPUTE_ID))

        # if no compute, use spare vThunder and prepare network for spare vThunder
        vthunder_for_amphora_subflow.add(a10_database_tasks.GetSpareComputeForProject(
            name=sf_name + '-' + a10constants.GET_SPARE_COMPUTE_FOR_PROJECT,
            requires=constants.COMPUTE_ID,
            provides=(constants.COMPUTE_ID, a10constants.SPARE_VTHUNDER)))
        vthunder_for_amphora_subflow.add(a10_network_tasks.PlugVipNetworkOnSpare(
            name=sf_name + '-' + a10constants.PLUG_VIP_NETWORK_ON_SPARE,
            requires=(a10constants.SPARE_VTHUNDER, constants.LOADBALANCER),
            provides=a10constants.ADDED_NETWORK))
        vthunder_for_amphora_subflow.add(vthunder_tasks.SparePostNetworkPlug(
            name=sf_name + '-' + a10constants.POST_SPARE_PLUG_NETWORK,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            requires=(a10constants.ADDED_NETWORK)))
        vthunder_for_amphora_subflow.add(a10_database_tasks.GetVThunderAmphora(
            name=sf_name + '-' + a10constants.GET_VTHUNDER_AMPHORA,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            provides=constants.AMPHORA))
        vthunder_for_amphora_subflow.add(vthunder_tasks.VThunderComputeConnectivityWait(
            name=sf_name + '-' + constants.AMP_COMPUTE_CONNECTIVITY_WAIT,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            requires=(constants.AMPHORA)))
        vthunder_for_amphora_subflow.add(vthunder_tasks.EnableInterfaceOnSpare(
            name=sf_name + '-' + a10constants.ENABLE_VTHUNDER_INTERFACE,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            requires=(a10constants.ADDED_NETWORK)))

        vthunder_for_amphora_subflow.add(database_tasks.UpdateAmphoraComputeId(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_COMPUTEID,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_ID)))
        vthunder_for_amphora_subflow.add(compute_tasks.ComputeActiveWait(
            name=sf_name + '-' + constants.COMPUTE_WAIT,
            requires=(constants.COMPUTE_ID, constants.AMPHORA_ID),
            provides=constants.COMPUTE_OBJ))
        vthunder_for_amphora_subflow.add(database_tasks.UpdateAmphoraInfo(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_INFO,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_OBJ),
            provides=constants.AMPHORA))
        # create vThunder entry in custom DB
        vthunder_for_amphora_subflow.add(a10_database_tasks.CreateVThunderEntry(
            name=sf_name + '-' + a10constants.CREATE_VTHUNDER_ENTRY,
            requires=(constants.AMPHORA, constants.LOADBALANCER),
            inject={"role": role, "status": constants.PENDING_CREATE}))
        # Get VThunder details from database
        vthunder_for_amphora_subflow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            name=sf_name + '-' + a10constants.VTHUNDER_BY_LB,
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        vthunder_for_amphora_subflow.add(database_tasks.ReloadLoadBalancer(
            name=sf_name + '-' + a10constants.RELOADLOAD_BALANCER,
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER))
        vthunder_for_amphora_subflow.add(a10_network_tasks.GetLBResourceSubnet(
            name=sf_name + '-' + a10constants.GET_LB_RESOURCE,
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.SUBNET))
        vthunder_for_amphora_subflow.add(
            a10_database_tasks.GetChildProjectsOfParentPartition(
                name=sf_name + '-' + a10constants.GET_PROJECT_COUNT,
                requires=[a10constants.VTHUNDER],
                rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
                provides=a10constants.PARTITION_PROJECT_LIST
            ))
        vthunder_for_amphora_subflow.add(
            a10_database_tasks.CountLoadbalancersInProjectBySubnet(
                name=sf_name + '-' + a10constants.GET_LB_COUNT_SUBNET,
                requires=[constants.SUBNET, a10constants.PARTITION_PROJECT_LIST],
                provides=a10constants.LB_COUNT_SUBNET))
        vthunder_for_amphora_subflow.add(a10_network_tasks.AllocateVIP(
            name=sf_name + '-' + a10constants.ALLOCATE_VIP,
            requires=[constants.LOADBALANCER, a10constants.LB_COUNT_SUBNET],
            provides=constants.VIP))
        vthunder_for_amphora_subflow.add(database_tasks.UpdateVIPAfterAllocation(
            name=sf_name + '-' + a10constants.UPDATE_VIP_AFTER_ALLOCATION,
            requires=(constants.LOADBALANCER_ID, constants.VIP),
            provides=constants.LOADBALANCER))
        vthunder_for_amphora_subflow.add(
            database_tasks.MarkAmphoraAllocatedInDB(
                name=sf_name + '-' + constants.MARK_AMPHORA_ALLOCATED_INDB,
                requires=(constants.AMPHORA, constants.LOADBALANCER_ID)))
        vthunder_for_amphora_subflow.add(database_tasks.ReloadAmphora(
            name=sf_name + '-' + constants.RELOAD_AMPHORA,
            requires=constants.AMPHORA_ID,
            provides=constants.AMPHORA))
        if role == constants.ROLE_MASTER:
            vthunder_for_amphora_subflow.add(database_tasks.MarkAmphoraMasterInDB(
                name=sf_name + '-' + constants.MARK_AMP_MASTER_INDB,
                requires=constants.AMPHORA))
            vthunder_for_amphora_subflow.add(
                vthunder_tasks.UpdateAcosVersionInVthunderEntry(
                    name=sf_name + '-' + a10constants.UPDATE_ACOS_VERSION_FOR_BACKUP_VTHUNDER,
                    requires=(a10constants.VTHUNDER)))
        elif role == constants.ROLE_BACKUP:
            vthunder_for_amphora_subflow.add(database_tasks.MarkAmphoraBackupInDB(
                name=sf_name + '-' + constants.MARK_AMP_BACKUP_INDB,
                requires=constants.AMPHORA))
            vthunder_for_amphora_subflow.add(
                a10_database_tasks.GetBackupVThunderByLoadBalancer(
                    name=sf_name + '-' + a10constants.BACKUP_VTHUNDER,
                    requires=constants.LOADBALANCER,
                    provides=a10constants.BACKUP_VTHUNDER))
            vthunder_for_amphora_subflow.add(
                vthunder_tasks.UpdateAcosVersionInVthunderEntry(
                    name=sf_name + '-' + a10constants.UPDATE_ACOS_VERSION_FOR_BACKUP_VTHUNDER,
                    rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}))
        elif role == constants.ROLE_STANDALONE:
            vthunder_for_amphora_subflow.add(
                database_tasks.MarkAmphoraStandAloneInDB(
                    name=sf_name + '-' + constants.MARK_AMP_STANDALONE_INDB,
                    requires=constants.AMPHORA))
            vthunder_for_amphora_subflow.add(
                vthunder_tasks.UpdateAcosVersionInVthunderEntry(
                    name=sf_name + '-' + a10constants.UPDATE_ACOS_VERSION_FOR_BACKUP_VTHUNDER,
                    requires=(a10constants.VTHUNDER)))

        # If spare vThunder is used, remove spare vThunder the database
        vthunder_for_amphora_subflow.add(a10_database_tasks.DeleteStaleSpareVThunder(
            name=sf_name + '-' + a10constants.DELETE_STALE_SPARE_VTHUNDER,
            requires=(a10constants.SPARE_VTHUNDER)))
        return vthunder_for_amphora_subflow

    def get_glm_license_subflow(self, prefix, role):
        sf_name = prefix + '-' + a10constants.ACTIVATE_GLM_LICENSE_SUBFLOW
        glm_license_subflow = linear_flow.Flow(sf_name)

        if role == constants.ROLE_BACKUP:
            glm_license_subflow.add(vthunder_tasks.SetVThunderHostname(
                name=sf_name + '-' + a10constants.SET_VTHUNDER_HOSTNAME,
                requires=constants.AMPHORA,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}
            ))
            glm_license_subflow.add(glm_tasks.ConfigureForwardProxyServer(
                name=sf_name + '-' + a10constants.CONFIGURE_PROXY_SERVER,
                requires=constants.FLAVOR,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}
            ))
            glm_license_subflow.add(glm_tasks.DNSConfiguration(
                name=sf_name + '-' + a10constants.CONFIGURE_DNS_NAMESERVERS,
                requires=constants.FLAVOR,
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
            ))
            glm_license_subflow.add(glm_tasks.ActivateFlexpoolLicense(
                name=sf_name + '-' + a10constants.ACTIVATE_FLEXPOOL_LICENSE,
                requires=(constants.AMPHORA, constants.FLAVOR),
                rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
            ))
        else:
            glm_license_subflow.add(vthunder_tasks.SetVThunderHostname(
                name=sf_name + '-' + a10constants.SET_VTHUNDER_HOSTNAME,
                requires=(constants.AMPHORA, a10constants.VTHUNDER)
            ))
            glm_license_subflow.add(glm_tasks.ConfigureForwardProxyServer(
                name=sf_name + '-' + a10constants.CONFIGURE_PROXY_SERVER,
                requires=(constants.FLAVOR, a10constants.VTHUNDER)
            ))
            glm_license_subflow.add(glm_tasks.DNSConfiguration(
                name=sf_name + '-' + a10constants.CONFIGURE_DNS_NAMESERVERS,
                requires=(constants.FLAVOR, a10constants.VTHUNDER)
            ))
            glm_license_subflow.add(glm_tasks.ActivateFlexpoolLicense(
                name=sf_name + '-' + a10constants.ACTIVATE_FLEXPOOL_LICENSE,
                requires=(constants.AMPHORA, a10constants.VTHUNDER, constants.FLAVOR),
            ))
        return glm_license_subflow

    def get_vrrp_subflow(self, prefix):
        sf_name = prefix + '-' + constants.GET_VRRP_SUBFLOW
        vrrp_subflow = linear_flow.Flow(sf_name)
        # TODO(omkartelee01) Need HA variables here
        vrrp_subflow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            name=sf_name + '-' + a10constants.GET_LOADBALANCER_FROM_DB,
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        vrrp_subflow.add(a10_database_tasks.GetBackupVThunderByLoadBalancer(
            name=sf_name + '-' + a10constants.GET_BACKUP_LOADBALANCER_FROM_DB,
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER),
            provides=a10constants.BACKUP_VTHUNDER))
        # Make sure devices are ready
        vrrp_subflow.add(vthunder_tasks.VThunderComputeConnectivityWait(
            name=sf_name + '-' + a10constants.WAIT_FOR_MASTER_SYNC + '-for-thunder',
            requires=(a10constants.VTHUNDER, constants.AMPHORA)))
        vrrp_subflow.add(vthunder_tasks.VThunderComputeConnectivityWait(
            name=sf_name + '-' + a10constants.WAIT_FOR_BACKUP_SYNC + '-for-thunder',
            rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
            requires=(constants.AMPHORA)))
        # VRRP Configuration
        vrrp_subflow.add(a10_database_tasks.AddProjectSetIdDB(
            name=sf_name + '-' + a10constants.ADD_VRRP_SET_ID_INDB,
            requires=constants.LOADBALANCER,
            provides=a10constants.SET_ID))
        vrrp_subflow.add(vthunder_tasks.ConfigureVRRPMaster(
            name=sf_name + '-' + a10constants.CONFIGURE_VRRP_FOR_MASTER_VTHUNDER,
            requires=(a10constants.VTHUNDER, a10constants.SET_ID)))
        vrrp_subflow.add(vthunder_tasks.ConfigureVRRPBackup(
            name=sf_name + '-' + a10constants.CONFIGURE_VRRP_FOR_BACKUP_VTHUNDER,
            requires=(a10constants.VTHUNDER, a10constants.SET_ID),
            rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}))
        vrrp_subflow.add(self._get_vrrp_status_subflow(sf_name))

        return vrrp_subflow

    def _get_vrrp_status_subflow(self, sf_name):
        check_vrrp_status = graph_flow.Flow(a10constants.CHECK_VRRP_STATUS)
        check_vrrp_status.add(vthunder_tasks.CheckVRRPStatus(
            name=a10constants.CHECK_VRRP_MASTER_STATUS,
            requires=(a10constants.VTHUNDER),
            provides=(a10constants.MASTER_VRRP_STATUS)))
        check_vrrp_status.add(vthunder_tasks.CheckVRRPStatus(
            name=a10constants.CHECK_VRRP_BACKUP_STATUS,
            rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER},
            provides=(a10constants.BACKUP_VRRP_STATUS)))
        confirm_vrrp = vthunder_tasks.ConfirmVRRPStatus(
            name=a10constants.CONFIRM_VRRP_STATUS,
            requires=(a10constants.MASTER_VRRP_STATUS, a10constants.BACKUP_VRRP_STATUS),
            provides=(a10constants.VRRP_STATUS))
        configure_vrrp = self._configure_vrrp_subflow(sf_name)
        check_vrrp_status.add(confirm_vrrp, configure_vrrp)
        check_vrrp_status.link(confirm_vrrp, configure_vrrp,
                               decider=self._is_vrrp_configured)
        return check_vrrp_status

    def _configure_vrrp_subflow(self, sf_name):
        configure_vrrp_subflow = linear_flow.Flow(sf_name)
        # Configure aVCS
        configure_vrrp_subflow.add(vthunder_tasks.ConfigureaVCSMaster(
            name=sf_name + '-' + a10constants.CONFIGURE_AVCS_SYNC_FOR_MASTER,
            requires=(a10constants.VTHUNDER)))
        configure_vrrp_subflow.add(vthunder_tasks.ConfigureaVCSBackup(
            name=sf_name + '-' + a10constants.CONFIGURE_AVCS_SYNC_FOR_BACKUP,
            rebind={a10constants.VTHUNDER: a10constants.BACKUP_VTHUNDER}))
        # Wait for aVCS sync
        configure_vrrp_subflow.add(vthunder_tasks.VCSSyncWait(
            name=sf_name + '-' + a10constants.VCS_SYNC_WAIT,
            requires=a10constants.VTHUNDER))
        return configure_vrrp_subflow

    def get_rack_vthunder_for_lb_subflow(
            self, vthunder_conf, prefix, role=constants.ROLE_STANDALONE):
        """ reload the loadbalancer and make entry in database"""

        sf_name = prefix + '-' + constants.GET_AMPHORA_FOR_LB_SUBFLOW

        amp_for_lb_flow = linear_flow.Flow(sf_name)
        amp_for_lb_flow.add(a10_database_tasks.CreateRackVthunderEntry(
            name=sf_name + '-' + 'create_rack_vThunder_entry_in_database',
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER_CONFIG)))
        amp_for_lb_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        return amp_for_lb_flow

    def _is_vrrp_configured(self, history):
        """Checks whether vrrp is configured
        :returns: True if vrrp is configured
        """
        return history[list(history.keys())[0]]

    def get_write_memory_flow(self, vthunder, store, deleteCompute):
        """Perform write memory for thunder """
        sf_name = 'a10-house-keeper' + '-' + a10constants.WRITE_MEMORY_THUNDER_FLOW

        write_memory_flow = linear_flow.Flow(sf_name)
        vthunder_store = {}
        vthunder_store[vthunder.vthunder_id] = vthunder
        write_memory_flow.add(a10_database_tasks.GetActiveLoadBalancersByThunder(
            requires=a10constants.VTHUNDER,
            rebind={a10constants.VTHUNDER: vthunder.vthunder_id},
            name='{flow}-{id}'.format(
                id=vthunder.vthunder_id,
                flow='GetActiveLoadBalancersByThunder'),
            provides=a10constants.LOADBALANCERS_LIST))
        write_memory_flow.add(a10_database_tasks.MarkLoadBalancersPendingUpdateInDB(
            name='{flow}-{id}'.format(
                id=vthunder.vthunder_id,
                flow='MarkLoadBalancersPendingUpdateInDB'),
            requires=a10constants.LOADBALANCERS_LIST))
        if not deleteCompute:
            write_memory_flow.add(vthunder_tasks.WriteMemoryHouseKeeper(
                requires=(a10constants.VTHUNDER, a10constants.LOADBALANCERS_LIST,
                          a10constants.WRITE_MEM_SHARED_PART),
                rebind={a10constants.VTHUNDER: vthunder.vthunder_id},
                name='{flow}-{partition}-{id}'.format(
                    id=vthunder.vthunder_id,
                    flow='WriteMemory-' + a10constants.WRITE_MEMORY_THUNDER_FLOW,
                    partition=a10constants.WRITE_MEM_FOR_SHARED_PARTITION),
                provides=a10constants.WRITE_MEM_SHARED))
            write_memory_flow.add(vthunder_tasks.WriteMemoryHouseKeeper(
                requires=(a10constants.VTHUNDER, a10constants.LOADBALANCERS_LIST),
                rebind={a10constants.VTHUNDER: vthunder.vthunder_id},
                name='{flow}-{partition}-{id}'.format(
                    id=vthunder.vthunder_id,
                    flow='WriteMemory-' + a10constants.WRITE_MEMORY_THUNDER_FLOW,
                    partition=a10constants.WRITE_MEM_FOR_LOCAL_PARTITION),
                provides=a10constants.WRITE_MEM_PRIVATE))
            write_memory_flow.add(a10_database_tasks.SetThunderLastWriteMem(
                requires=(a10constants.VTHUNDER,
                          a10constants.WRITE_MEM_SHARED,
                          a10constants.WRITE_MEM_PRIVATE),
                rebind={a10constants.VTHUNDER: vthunder.vthunder_id},
                name='{flow}-{id}'.format(
                    id=vthunder.vthunder_id,
                    flow='SetThunderLastWriteMem')))
        else:
            write_memory_flow.add(a10_database_tasks.SetThunderLastWriteMem(
                requires=(a10constants.VTHUNDER),
                inject={a10constants.WRITE_MEM_SHARED: True, a10constants.WRITE_MEM_PRIVATE: True},
                rebind={a10constants.VTHUNDER: vthunder.vthunder_id},
                name='{flow}-{id}'.format(
                    id=vthunder.vthunder_id,
                    flow='SetThunderLastWriteMem')))
        write_memory_flow.add(a10_database_tasks.MarkLoadBalancersActiveInDB(
            name='{flow}-{id}'.format(
                id=vthunder.vthunder_id,
                flow='MarkLoadBalancersActiveInDB'),
            requires=a10constants.LOADBALANCERS_LIST))

        store.update(vthunder_store)
        return write_memory_flow

    def get_reload_check_flow(self, vthunder, store):
        """Perform write memory for thunder """
        sf_name = 'a10-house-keeper' + '-' + a10constants.RELOAD_CHECK_THUNDER_FLOW

        reload_check_flow = linear_flow.Flow(sf_name)
        vthunder_store = {}
        vthunder_store[vthunder.vthunder_id] = vthunder
        reload_check_flow.add(a10_database_tasks.GetActiveLoadBalancersByThunder(
            requires=a10constants.VTHUNDER,
            rebind={a10constants.VTHUNDER: vthunder.vthunder_id},
            name='{flow}-{id}'.format(
                id=vthunder.vthunder_id,
                flow='GetActiveLoadBalancersByThunder'),
            provides=a10constants.LOADBALANCERS_LIST))
        reload_check_flow.add(vthunder_tasks.WriteMemoryThunderStatusCheck(
            name='{flow}-{id}'.format(
                id=vthunder.vthunder_id,
                flow='WriteMemoryThunderStatusCheck'),
            requires=(a10constants.VTHUNDER, a10constants.LOADBALANCERS_LIST),
            rebind={a10constants.VTHUNDER: vthunder.vthunder_id}))

        store.update(vthunder_store)
        return reload_check_flow

    def get_failover_spare_vthunder_flow(self):
        """Perform failover for spare vthunder"""
        sf_name = 'a10-house-keeper-failover-spare-vthunder'
        failover_flow = linear_flow.Flow(sf_name)

        failover_flow.add(a10_database_tasks.GetVThunderAmphora(
            name=sf_name + '-' + a10constants.GET_VTHUNDER_AMPHORA,
            requires=a10constants.VTHUNDER,
            provides=constants.AMPHORA))
        failover_flow.add(compute.ComputeDelete(
            name=sf_name + '-' + a10constants.COMPUTE_DELETE,
            requires=constants.AMPHORA))
        failover_flow.add(a10_database_tasks.DeleteStaleSpareVThunder(
            name=sf_name + '-' + a10constants.DELETE_STALE_SPARE_VTHUNDER,
            rebind={a10constants.SPARE_VTHUNDER: a10constants.VTHUNDER}))

        return failover_flow

    def _get_failover_use_spare_amphora_subflow(self, prefix):
        """Flow to get spare amphora for failvoer"""
        sf_name = 'failover_get_spare_amphora'
        get_spare_flow = linear_flow.Flow(sf_name)
        get_spare_flow.add(a10_network_tasks.PlugNetworksByID(
            name=sf_name + '-' + a10constants.PLUG_NETWORK_BY_IDS,
            requires=(a10constants.NETWORK_LIST),
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            provides=a10constants.ADDED_NETWORK))
        get_spare_flow.add(vthunder_tasks.SparePostNetworkPlug(
            name=sf_name + '-' + a10constants.POST_SPARE_PLUG_NETWORK,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            requires=(a10constants.ADDED_NETWORK)))
        get_spare_flow.add(a10_database_tasks.GetVThunderAmphora(
            name=sf_name + '-' + a10constants.GET_VTHUNDER_AMPHORA,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            provides=constants.AMPHORA))
        get_spare_flow.add(vthunder_tasks.VThunderComputeConnectivityWait(
            name=sf_name + '-' + constants.AMP_COMPUTE_CONNECTIVITY_WAIT,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            requires=(constants.AMPHORA)))
        get_spare_flow.add(vthunder_tasks.EnableInterfaceOnSpare(
            name=sf_name + '-' + a10constants.ENABLE_VTHUNDER_INTERFACE,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            requires=(a10constants.ADDED_NETWORK)))

        return get_spare_flow

    def _get_failover_create_amphora_subflow(self, prefix):
        """Flow to create amphora for failover"""
        create_amp_flow = linear_flow.Flow(constants.CREATE_AMPHORA_FLOW)
        sf_name = prefix + '-' + a10constants.SPARE_VTHUNDER_CREATE

        create_amp_flow.add(database_tasks.CreateAmphoraInDB(
            name=sf_name + '-' + constants.CREATE_AMPHORA_INDB,
            provides=constants.AMPHORA_ID))
        create_amp_flow.add(compute_tasks.ComputeCreate(
            name=sf_name + '-' + constants.COMPUTE_CREATE,
            requires=(constants.AMPHORA_ID, a10constants.NETWORK_LIST),
            provides=constants.COMPUTE_ID))
        create_amp_flow.add(database_tasks.UpdateAmphoraComputeId(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_COMPUTEID,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_ID)))

        create_amp_flow.add(database_tasks.MarkAmphoraBootingInDB(
            name=sf_name + '-' + constants.MARK_AMPHORA_BOOTING_INDB,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_ID)))
        create_amp_flow.add(compute_tasks.ComputeActiveWait(
            name=sf_name + '-' + constants.COMPUTE_WAIT,
            requires=(constants.COMPUTE_ID, constants.AMPHORA_ID),
            provides=constants.COMPUTE_OBJ))
        create_amp_flow.add(database_tasks.UpdateAmphoraInfo(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_INFO,
            requires=(constants.AMPHORA_ID, constants.COMPUTE_OBJ),
            provides=constants.AMPHORA))
        create_amp_flow.add(a10_database_tasks.CreateSpareVThunderEntry(
            requires=(constants.AMPHORA),
            provides=a10constants.SPARE_VTHUNDER))
        create_amp_flow.add(
            vthunder_tasks.VThunderComputeConnectivityWait(
                name=sf_name + '-' + constants.AMP_COMPUTE_CONNECTIVITY_WAIT,
                requires=constants.AMPHORA,
                rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER}))
        create_amp_flow.add(a10_database_tasks.MarkVThunderStatusInDB(
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            inject={"status": a10constants.READY}))
        create_amp_flow.add(
            vthunder_tasks.UpdateAcosVersionInVthunderEntry(
                name=sf_name + '-' + a10constants.UPDATE_ACOS_VERSION_IN_VTHUNDER_ENTRY,
                rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER}))
        create_amp_flow.add(vthunder_tasks.EnableInterfaceOnSpare(
            name=sf_name + '-' + a10constants.ENABLE_VTHUNDER_INTERFACE,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER}))
        return create_amp_flow

    def _get_failover_amphora_subflow(self, prefix):
        """Flow to get vThunder for failover process"""
        sf_name = prefix + '-' + a10constants.GET_AMPHORA_FOR_FAILOVER
        failover_get_amp_flow = graph_flow.Flow(sf_name)

        check_spare = a10_database_tasks.TryGetSpareCompute(
            name=sf_name + '-' + 'check-spare-exist',
            provides=(a10constants.SPARE_VTHUNDER))
        get_spare_amp = self._get_failover_use_spare_amphora_subflow(sf_name)
        create_amp = self._get_failover_create_amphora_subflow(sf_name)
        failover_get_amp_flow.add(check_spare, get_spare_amp, create_amp)
        failover_get_amp_flow.link(check_spare, get_spare_amp,
                                   decider=self._allocate_amp_to_lb_decider,
                                   decider_depth='flow')
        failover_get_amp_flow.link(check_spare, create_amp,
                                   decider=self._create_new_amp_for_lb_decider,
                                   decider_depth='flow')
        return failover_get_amp_flow

    def _get_failover_configure_vcs_subflow(self, prefix):
        """Flow to get failover vcs configure process"""
        sf_name = prefix + '-' + 'conf-vcs'
        failover_vcs_flow = linear_flow.Flow(sf_name)

        failover_vcs_flow.add(a10_database_tasks.GetProjectVRRPSetId(
            name=sf_name + '-' + a10constants.GET_VRRP_SET_ID_INDB,
            requires=a10constants.VTHUNDER,
            provides=a10constants.SET_ID))
        failover_vcs_flow.add(a10_database_tasks.GetVThunderDeviceID(
            name=sf_name + '-' + a10constants.GET_VCS_DEVICE_ID,
            requires=a10constants.VTHUNDER,
            provides=a10constants.DEVICE_ID))
        failover_vcs_flow.add(vthunder_tasks.ConfigureVRRPFailover(
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            requires=(a10constants.DEVICE_ID, a10constants.SET_ID)))
        failover_vcs_flow.add(vthunder_tasks.ConfigureaVCSFailover(
            name=sf_name + '-' + a10constants.CONFIGURE_AVCS_FOR_FAILOVER,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER},
            requires=a10constants.DEVICE_ID))
        # Wait for aVCS sync
        failover_vcs_flow.add(vthunder_tasks.VCSSyncWait(
            name=sf_name + '-' + a10constants.VCS_SYNC_WAIT,
            rebind={a10constants.VTHUNDER: a10constants.SPARE_VTHUNDER}))

        return failover_vcs_flow

    def get_failover_vcs_vthunder_flow(self):
        """Perform failover for VCS vthunder device"""
        sf_name = 'a10-house-keeper-failover-vcs-vthunder'
        failover_flow = linear_flow.Flow(sf_name)

        failover_flow.add(a10_database_tasks.GetComputeVThundersAndLoadBalancers(
            name=sf_name + '-' + a10constants.GET_LBS_BY_THUNDER,
            requires=a10constants.VTHUNDER,
            provides=(a10constants.VTHUNDER_LIST, a10constants.LOADBALANCERS_LIST)))
        # Switch role may not necessary
        failover_flow.add(a10_database_tasks.SetVThunderToStandby(
            name=sf_name + '-' + a10constants.SET_VTHUNDER_TO_STANDBY,
            requires=a10constants.VTHUNDER_LIST))
        failover_flow.add(a10_database_tasks.LoadBalancerListToErrorOnRevertTask(
            name=sf_name + '-' + a10constants.MARK_LB_LIST_ERROR_ON_REVERT,
            requires=a10constants.LOADBALANCERS_LIST))
        failover_flow.add(a10_database_tasks.MarkLoadBalancersPendingUpdateInDB(
            name=sf_name + '-' + a10constants.MARK_LB_PENIND_UPDATE_IN_DB,
            requires=a10constants.LOADBALANCERS_LIST))
        failover_flow.add(a10_network_tasks.GetVThunderNetworkList(
            name=sf_name + '-' + a10constants.GET_VTHUNDER_NETWORK_LIST,
            requires=a10constants.VTHUNDER,
            provides=a10constants.NETWORK_LIST))
        failover_flow.add(self._get_failover_amphora_subflow(sf_name))
        failover_flow.add(self._get_failover_configure_vcs_subflow(sf_name))

        failover_flow.add(compute_tasks.DeleteStaleCompute(
            name=sf_name + '-' + 'delete-stale-compute',
            requires=(a10constants.VTHUNDER)))
        failover_flow.add(a10_database_tasks.FailoverPostDbUpdate(
            name=sf_name + '-' + a10constants.POST_FAILOVER_DB_UPDATE,
            requires=(a10constants.VTHUNDER, a10constants.SPARE_VTHUNDER)))
        failover_flow.add(a10_database_tasks.DeleteStaleSpareVThunder(
            name=sf_name + '-' + a10constants.DELETE_STALE_SPARE_VTHUNDER,
            requires=(a10constants.SPARE_VTHUNDER)))
        failover_flow.add(a10_database_tasks.MarkLoadBalancersActiveInDB(
            name=sf_name + '-' + a10constants.MARK_LB_ACTIVE_IN_DB,
            requires=a10constants.LOADBALANCERS_LIST))
        return failover_flow

    def get_failover_restore_vthunder_flow(self):
        """Perform vthunder restore"""
        sf_name = 'a10-house-keeper-failover-restore-vthunder'
        failover_flow = linear_flow.Flow(sf_name)

        return failover_flow
